/**
 * @file FuzzyClearBall.cc
 * Implementation of the FuzzyClearBall skill.
 *
 * @par Unit:
 *      Framework/skill
 * $Revision: 1.5 $
 * $Author: andbo928 $
 * $Date: 2001/04/23 10:18:46 $
 */

#include "FuzzyClearBall.h"
#include "KickCommand.h"
#include "KickBallToSpeedVector.h"
#include "Line.h"

#define KICK_RANGE 30
#define LOOK_RANGE 20
#define SAFE_MARGIN 2
#define CB_FRIENDLY 0.2
#define CB_OPPONENT -1.0
#define CB_ADVANCE  0.5
#define CB_PRECISION 0.0

RS_BEGIN_NAMESPACE

// --- CONSTRUCTORS ---

/**
 * Constructor.
 */

FuzzyClearBall::FuzzyClearBall()
  : Skill(),
    parameterView(NULL), //,
    markerView(NULL),
    //    agentView(NULL),
    worldStateView(NULL),
    kickBall(Vector())
{
  getView(parameterView, PARAMETER_VIEW_ID);
  getView(markerView, MARKER_VIEW_ID, parameterView);
  //  getView(agentView, AGENT_VIEW_ID, parameterView);
  getView(worldStateView, WORLD_STATE_VIEW_ID);
}

// --- DESTRUCTOR ---

FuzzyClearBall::~FuzzyClearBall()
{
  viewManager->releaseView(PARAMETER_VIEW_ID);
  //  viewManager->releaseView(MARKER_VIEW_ID);
  //  viewManager->releaseView(AGENT_VIEW_ID);
  viewManager->releaseView(WORLD_STATE_VIEW_ID);
}


// --- METHODS ---

/**
 * Genrates a plan for achieving a pass.
 * Chooses the closest of the passvictims if the argument is
 * supplied, otherwise pass the best player according to 
 * TerritoryControlView
 */

Plan FuzzyClearBall::generatePlan()
{
  Vector passVector = calculateClearVector();
  if(kickBall.applicable(passVector)) {
    return kickBall.generatePlan(passVector); 
  }
  return Plan();
}

Vector FuzzyClearBall::calculateClearVector() {

  // Construct the edge lines.
  TVector ltFlagPos = TVector(markerView->getMarker(FLAG_LT, NOW).getPosition());
  ltFlagPos = ltFlagPos + TVector(SAFE_MARGIN, -SAFE_MARGIN);
  TVector rtFlagPos = TVector(markerView->getMarker(FLAG_RT, NOW).getPosition());
  rtFlagPos = rtFlagPos + TVector(-SAFE_MARGIN, -SAFE_MARGIN);
  TVector lbFlagPos = TVector(markerView->getMarker(FLAG_LB, NOW).getPosition());
  lbFlagPos = lbFlagPos + TVector(SAFE_MARGIN, SAFE_MARGIN);
  TVector rbFlagPos = TVector(markerView->getMarker(FLAG_RB, NOW).getPosition());
  rbFlagPos = rbFlagPos + TVector(-SAFE_MARGIN, SAFE_MARGIN);

  TVector horiz = TVector::TPolarVector(1,0);

  const WorldState& worldState = worldStateView->getWorldState();
  TVector agentPos = worldState.getAgent().getPosition();
  TLine topLimit(agentPos, TVector::TPolarVector(-1,60));
  TLine botLimit(agentPos, TVector::TPolarVector(1,60));

  TLine topLtoR(ltFlagPos, horiz);
  TLine botLtoR(lbFlagPos, horiz);

  topLimit.extendToTLine(topLtoR);
  botLimit.extendToTLine(botLtoR);

//    Point topP = topLimit.getPoint() + topLimit.getVector();
//    Point botP = botLimit.getPoint() + botLimit.getVector();
  //  cout << "Tp: " << topP << " Bp: " << botP << endl;

  topLtoR.setVector(topLimit.getEnd() - topLtoR.getStart());
  botLtoR.setVector(botLimit.getEnd() - botLtoR.getStart());

  topLtoR.extendToCircle(agentPos, KICK_RANGE);
  botLtoR.extendToCircle(agentPos, KICK_RANGE);

  TVector topV = topLtoR.getEnd() - agentPos;
  TVector botV = botLtoR.getEnd() - agentPos;

  if (topV.getAngle() < -60) {
    topV.setAngle(-60);
  }

  if (botV.getAngle() > 60) {
    botV.setAngle(60);
  }

//   cout << "Top: " << topV.getAngle().getDeg() << " Bot: " << botV.getAngle().getDeg() << endl;

  /*  AngleDeg dir = territoryControlView->getMaxMinNoCache(CB_FRIENDLY, 
							CB_OPPONENT, 
							CB_ADVANCE, 
							CB_PRECISION,
							clearSafeness, 
							topV.getAngle(),
							botV.getAngle());

   cout << "Cleardir: " << dir << endl;

  if(agentView->getBodyDirection(NOW).isKnown()) {
    dir = dir - agentView->getBodyDirection(NOW).getAngle().getAngle().getAngle();
  }
  */

  Float unsafeDirs[128];
  int numUnsafeDirs = 0;

  Float minAng = topV.getAngle();
  Float maxAng = botV.getAngle();

  //  cout << "limits: " << minAng << ", " << maxAng << endl;
  
  unsafeDirs[numUnsafeDirs] = minAng;
  numUnsafeDirs++;
  unsafeDirs[numUnsafeDirs] = maxAng;
  numUnsafeDirs++;

  const std::vector<PlayerState>& oppVec = worldState.getOpponents();
  for(unsigned int i = 0; i < oppVec.size(); i++) {
    TVector plVec = oppVec[i].getPosition() - agentPos;
    if(plVec.getC() > 0.01 &&
       plVec.getLength() < LOOK_RANGE) {
      Float curAng = plVec.getAngle();
      if(curAng > minAng && curAng < maxAng) {
	unsafeDirs[numUnsafeDirs] = curAng;
	numUnsafeDirs++;
      }
    }
  }

//   oppVec = playerView->getUnknownPlayers(NOW);
//   for(unsigned int i = 0; i < oppVec.size(); i++) {
//     VectorUCO plVec = VectorUCO(oppVec[i].getPos() - agentView->getAgentPosition(NOW));
//     if(plVec.getLength().getConfidence() > 0.01 &&
//        plVec.getLength() > LOOK_RANGE) {
//       Float curAng = plVec.getAngle().getDeg();
//       if(curAng > minAng && curAng < maxAng) {
// 	unsafeDirs[numUnsafeDirs] = curAng;
// 	numUnsafeDirs++;
//       }
//     }
//   }

  // Sort the angles!
  // Bubbles are beutiful!

  bool sorted = false;
  Float tmp;
  while(!sorted) {
    sorted = true;
    for(int i = 0; i < numUnsafeDirs-1; i++) {
      if(unsafeDirs[i] > unsafeDirs[i+1]) {
	tmp = unsafeDirs[i];
	unsafeDirs[i] = unsafeDirs[i+1];
	unsafeDirs[i+1] = tmp;
	sorted = false;
      }
    }
  }

  /*  cout << "Unsafe: [";
  for(int i = 0; i < numUnsafeDirs; i++)
    cout << unsafeDirs[i] << ",";

    cout << "]" << endl; */

  // Calc the largest interval.

  Float maxStart = 0, maxEnd, maxVal;
  maxVal = -1;

  for(int i = 0; i < numUnsafeDirs-1; i++) {
    Float curVal = unsafeDirs[i+1] - unsafeDirs[i];
    if(curVal > maxVal) {
      maxStart = unsafeDirs[i];
      maxEnd = unsafeDirs[i+1];
      maxVal = curVal;
    }
  }

  // WhoppidooooO!

  //  cout << "Wanted dir" << (maxStart + maxVal / 2 ) << endl;

  AngleDeg dir = AngleDeg(maxStart + maxVal / 2);
//  			  - agentView->getBodyDirection(NOW).getAngle().getAngle().getAngle().getDeg());

  // KickBall is relative to the player.

  return Vector(dir, KICK_RANGE);
}


/**
 * Persistent.
 * FALSE
 */
FuzzyBool FuzzyClearBall::persistent()
{
  return kickBall.persistent();
}

/**
 * Returns the applicable value
 */

FuzzyBool FuzzyClearBall::applicable()
{
  const WorldState worldState = worldStateView->getWorldState();
  if(!worldState.getAgent().isBallKickable(worldState.getBall()))
    return FALSE;

  Vector clearVector = calculateClearVector();
  if(clearVector.getLength() == 0)
    return FALSE;
  if(kickBall.applicable(clearVector)) {
//     cout << "CS: " << clearSafeness << endl;
    //    return FuzzyBool(minMax(0.0, clearSafeness + 1, 1.0));
    return FuzzyBool(1.0);
  }
  return FALSE;
}

RS_END_NAMESPACE
