/* -*- Mode: C++ -*-
 * $Id: KickBallToSpeedVector.cc,v 1.11 2001/07/28 15:29:20 andbo928 Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: KickBallToSpeedVector.cc
 * Unitname: Framework/skill
 * $Revision: 1.11 $
 * Created by: Fredrik Heintz 1999-xx-xx
 * Last modified by $Author: andbo928 $ $Date: 2001/07/28 15:29:20 $
 *
 *
 * HISTORY
 *
 * $Log: KickBallToSpeedVector.cc,v $
 * Revision 1.11  2001/07/28 15:29:20  andbo928
 * Despammed.
 *
 * Revision 1.10  2001/07/28 15:28:49  andbo928
 * Changed buffer.
 *
 * Revision 1.9  2001/07/28 13:32:34  mikbr212
 * Removed some view quality stuff
 *
 * Revision 1.8  2001/07/01 12:36:46  andbo928
 * New implementation due to server changes.
 *
 * Revision 1.7  2001/04/24 23:04:56  sigge
 * InPenaltyArea is now fuzzy.
 *
 * Revision 1.6  2001/04/24 14:09:44  sigge
 * Added emergency predicate to decide when to use one kick only.
 *
 * Revision 1.5  2001/04/23 22:58:45  mikbr212
 * added setC(1) to states
 *
 * Revision 1.4  2001/04/20 19:01:13  mikbr212
 * added default constructor
 *
 * Revision 1.3  2001/04/14 19:26:08  mikbr212
 * Uses a single plan. Works good.
 *
 * Revision 1.2  2001/04/14 19:10:46  mikbr212
 * Bad skill which refreshes the ballState every cycle
 *
 * Revision 1.1.1.1  2001/04/08 11:33:58  sigge
 * Imported sources
 *
 * Revision 1.8  2000/11/08 19:32:07  mikbr212
 * Fixed mem-leak in generatePlan
 *
 * Revision 1.7  2000/11/07 18:10:43  mikbr212
 * changed persistent
 *
 * Revision 1.6  2000/10/30 18:32:45  mikbr212
 * added comments
 *
 */

/**
 * @file KickBallToSpeedVector.cc
 * Implementation file for the skill class KickBallToSpeedVector.
 *
 * @par Unit:
 *      Framework/skill
 * $Revision: 1.11 $
 * $Author: andbo928 $
 * $Date: 2001/07/28 15:29:20 $
 */


#include "KickBallToSpeedVector.h"
#include "InPenaltyArea.h"
#include <vector>

RS_BEGIN_NAMESPACE

KickBallToSpeedVector::KickBallToSpeedVector()
  : vector(0,0),
    parameterView(NULL),
    worldStateView(NULL)
{
  getView(parameterView, PARAMETER_VIEW_ID);
  getView(worldStateView, WORLD_STATE_VIEW_ID);
  computePositions();
} 

/**
 * Constructor.
 */
KickBallToSpeedVector::KickBallToSpeedVector(const TVector& vec)
  : vector(vec),
    parameterView(NULL),
    worldStateView(NULL)
{
  getView(parameterView, PARAMETER_VIEW_ID);
  getView(worldStateView, WORLD_STATE_VIEW_ID);
  computePositions();
} 

/**
 * Destructor.
 */
KickBallToSpeedVector::~KickBallToSpeedVector()
{
  viewManager->releaseView(PARAMETER_VIEW_ID);
  viewManager->releaseView(WORLD_STATE_VIEW_ID);
}

/**
 * Sets vector.
 */
void KickBallToSpeedVector::setVector(const TVector& vec) {
  vector = vec;
  computePositions();
}

/**
 * Retrieves vector.
 */
TVector KickBallToSpeedVector::getVector() const {
  return vector;
}

/**
 * Generates a plan.
 */
Plan KickBallToSpeedVector::generatePlan(const TVector& vec) {
  setVector(vec);
  return generatePlan();
}

void KickBallToSpeedVector::computePositions() {
  positions.clear();

  const double avgKick = 0.75; 

  double deltaX = 100 / 2 * parameterView->SP_kick_power_rate * avgKick;
  double deltaY = parameterView->SP_player_size + parameterView->SP_ball_size + 0.1;
  double kickR = parameterView->SP_kickable_area * 0.9;
  
  if(deltaX*deltaX + deltaY*deltaY > kickR*kickR)
    deltaX = sqrt(kickR*kickR - deltaY*deltaY);

  TVector desBallPosR(-deltaX,deltaY);
  TVector desBallPosL(-deltaX,-deltaY);

  TVector desBallPosR2(deltaX,deltaY);
  TVector desBallPosL2(deltaX,-deltaY);

  desBallPosR.rotate(vector.getAngle());
  desBallPosL.rotate(vector.getAngle());

  desBallPosR2.rotate(vector.getAngle());
  desBallPosL2.rotate(vector.getAngle());
  
  positions.push_back(desBallPosR);
  positions.push_back(desBallPosR2);
  positions.push_back(desBallPosL);
  positions.push_back(desBallPosL2);

}

/**
 * Generates a plan.
 */
Plan KickBallToSpeedVector::generatePlan() {

  const float BUFFER(0.3);
  const float DIST(0.5 * (parameterView->SP_kickable_area - parameterView->SP_player_size) + parameterView->SP_player_size);

  const BallState& ball = worldStateView->getWorldState().getBall();
  const AgentState& agent = worldStateView->getWorldState().getAgent();

  if(!applicable())
    return Plan();

  TVector b2a = agent.getPosition() - ball.getPosition();
  TVector nv = vector.normalized();

  TVector shortest = b2a - (b2a * nv) * nv;

  // Try to kick it!
  WorldState ws = worldStateView->getWorldState();

  ws = worldStateView->getWorldState();
  if(shortest.getLength() < (parameterView->SP_player_size + BUFFER)
     && (vector * b2a) > 0.0)
  {
    // Ball will collide with agent.
    // Try to kick it to the left or right of the agent.
    // Relative the desired vector.
    // Choose the position that is closest.
    
    TVector r = nv;
    r.rotate(90);

    TVector p1 = agent.getPosition() + r * DIST;
    TVector p2 = agent.getPosition() - r * DIST;

    TVector delta1 = p1 - ball.getPosition();
    TVector delta2 = p2 - ball.getPosition();

    if(delta1.getLength() < delta2.getLength())
    {
      // Right
      if(ws.doKickToPosition(p1))
      {
//  	cout << "Right." << endl;
	ws.doTurnNeckTowardsBall();
	ws.doChangeViewNarrow();
	ws.doChangeViewHigh();
	return Plan(ws.getStep());
      }
    }
    else
    {
      // Left
      if(ws.doKickToPosition(p2))
      {
//  	cout << "Left." << endl;
	ws.doTurnNeckTowardsBall();
	ws.doChangeViewNarrow();
	ws.doChangeViewHigh();
	return Plan(ws.getStep());
      }
    }

    // If it's moving fast Try to stop it.
    if(ball.getSpeed().getLength() > 0.1)
    {
      ws = worldStateView->getWorldState();
      if(ws.doKickToSpeed(TVector(0,0)))
      {
//  	cout << "Stop!" << endl;
	ws.doTurnNeckTowardsBall();
	ws.doChangeViewNarrow();
	ws.doChangeViewHigh();
	return Plan(ws.getStep());
      }
    }
  }

  // Try to kick it!
  ws = worldStateView->getWorldState();
  if(ws.doKickToSpeed(vector))
  {
//      cout << "Final!" << endl;
    ws.doTurnNeckTowardsBall();
    ws.doChangeViewNormal();
    ws.doChangeViewHigh();
    return Plan(ws.getStep());
  }

  // If it's moving fast Try to stop it.
  if(ball.getSpeed().getLength() > 0.1)
  {
    ws = worldStateView->getWorldState();
    if(ws.doKickToSpeed(TVector(0,0)))
    {
//        cout << "Stop!" << endl;
      ws.doTurnNeckTowardsBall();
      ws.doChangeViewNarrow();
      ws.doChangeViewHigh();
      return Plan(ws.getStep());
    }
  }

  // Turn towards the ball if needed.
  TAngle dir = agent.getDirection();
  TAngle dir2 = (-b2a).getAngle();
  if(fabs(dir - dir2) > 10)
  {
    cout << fabs(dir - dir2) << endl;
    ws = worldStateView->getWorldState();
    if(ws.doTurnTowardsBall())
    {
//        cout << "Turning!" << endl;
      ws.doTurnNeckTowardsBall();
      ws.doChangeViewNarrow();
      ws.doChangeViewHigh();
      return Plan(ws.getStep());
    }
  }

//    cout << "BadFinal!" << endl;
  ws = worldStateView->getWorldState();
  ws.doKickToSpeed(vector);
  ws.doTurnNeckTowardsBall();
  ws.doChangeViewNormal();
  ws.doChangeViewHigh();
  return Plan(ws.getStep());

//    // The simple way!
//    ws = worldStateView->getWorldState();
//    if(ws.doKickToSpeed(vector))
//    {
//      cout << "Final!" << endl;
//      ws.doTurnNeckTowardsBall();
//      return Plan(ws.getStep());
//    }

//    // One kick is'nt enough.
//    // Try to kick it in front of us.
//    ws = worldStateView->getWorldState();
//    if( ws.doKickToPosition(agent.getPosition()
//  			  + agent.getDirectionVector() * 0.5) )
//    {
//      cout << "Front!" << endl;
//      ws.doTurnNeckTowardsBall();
//      return Plan(ws.getStep());
//    }

//    // One kick is'nt enough.
//    // Try to kick it to the left of us.
//    ws = worldStateView->getWorldState();
//    if( ws.doKickToPosition(agent.getPosition()
//  			  + TVector::TPolarVector(0.5, agent.getDirection() - 90)))
//    {
//      cout << "Left!" << endl;
//      ws.doTurnNeckTowardsBall();
//      return Plan(ws.getStep());
//    }

//    // Just try to stop the ball.
//    ws = worldStateView->getWorldState();
//    ws.doKickToSpeed(TVector(0,0));
//    ws.doTurnNeckTowardsBall();
//    return Plan(ws.getStep());
 
//    { // Try if a single kick is enough
//      BallState newBall(ball);
//      AgentState newAgent(agent);
//      if(newAgent.doKickToPosition(vector, newBall) || emergency()) {
//        Step step(newAgent.getStep());
//        newBall.step();
//        newAgent.step();
//        if(!newAgent.overlaps(newBall)) {
//  //  	cout << "Single kick" << endl;
//  	return Plan(step);
//        }
//      }
//    }
  
//    Plan plan;
//    BallState newBall(ball);
//    AgentState newAgent(agent);
  
//    int rightLeft = 0;
//    double distR = (ball.getPosition() - agent.getPosition() - positions[0]).getLength();
//    double distL = (ball.getPosition() - agent.getPosition() - positions[2]).getLength();
//    if(distL < distR)
//      rightLeft = 2;

//    // Till bakre positionen
//    for(int i=0; i<3; i++) {
//      newAgent.setC(1);
//      newBall.setC(1);
//      newAgent.doKickToPosition(positions[0+rightLeft] + newAgent.getPosition(), newBall);
//      plan.addStep(newAgent.getStep());
//      newBall.step();
//      newAgent.step();
//    }
//    // framspark
//    newAgent.setC(1);
//    newBall.setC(1);
//    newAgent.doKickToPosition(positions[1+rightLeft] + newAgent.getPosition(), newBall);
//    plan.addStep(newAgent.getStep());
//    newBall.step();
//    newAgent.step();
//    // sista sparken
//    newAgent.setC(1);
//    newBall.setC(1);
//    newAgent.doKickToSpeed(vector,newBall);
//    plan.addStep(newAgent.getStep());
//    newBall.step();
//    newAgent.step();
  
//  //    if(rightLeft == 2) 
//  //      cout << "Back kick from right. Final speed: " <<  + newBall.getSpeed().getLength() << endl;
//  //    else
//  //      cout << "Back kick from left. Final speed: " <<  + newBall.getSpeed().getLength() << endl;

//    return plan;
}

/**
 * Checks if this skill should be called again (this cycle).
 */
FuzzyBool KickBallToSpeedVector::persistent()
{
  return FALSE; // This is a one-step skill
}

/**
 * Checks if the preconds are true.
 */
FuzzyBool KickBallToSpeedVector::applicable(const TVector& vec) {
  setVector(vec);
  return applicable();
}

/**
 * Checks if the preconds are true.
 */
FuzzyBool KickBallToSpeedVector::applicable() {
  const AgentState& agent = worldStateView->getWorldState().getAgent();
  const BallState& ball = worldStateView->getWorldState().getBall();
  return ball.getC() > 0.4 && agent.getKickPowerRate(ball) != 0;
}





RS_END_NAMESPACE
