/* -*- Mode: C++ -*-
 * $Id: MarkPlayer.cc,v 1.9 2001/07/11 16:35:50 sigge Exp $
 *
 * IDENTIFICATION
 *
 * Filename: MarkPlayer.h
 * Unitname: Framework/view
 * $Revision: 1.9 $
 * Created by: Fredrik Heintz 2000-09-12
 * Last modified by $Author: sigge $ $Date: 2001/07/11 16:35:50 $
 *
 *
 * HISTORY
 *
 * $Log: MarkPlayer.cc,v $
 * Revision 1.9  2001/07/11 16:35:50  sigge
 * Modified Monitor protocol for size and decoding speed.
 *
 * Revision 1.8  2001/04/24 21:39:57  mikbr212
 * added avoid-offside
 *
 * Revision 1.7  2001/04/24 21:25:33  mikbr212
 * fixed bug
 *
 * Revision 1.6  2001/04/24 20:50:50  mikbr212
 * ??
 *
 * Revision 1.5  2001/04/24 19:12:28  mikbr212
 * added sticky
 *
 * Revision 1.4  2001/04/24 14:10:12  sigge
 * Mark players 3 meters ahead of them.
 *
 * Revision 1.3  2001/04/24 12:30:47  mikbr212
 * fixed
 *
 * Revision 1.2  2001/04/23 22:57:11  mikbr212
 * Killed and made a new skill. Not done yet
 *
 *
 */

/**
 * @file MarkPlayer.cc
 * Implementation of the MarkPlayer skill.
 *
 * @par Unit:
 *      Framework/skill
 * $Revision: 1.9 $
 * $Author: sigge $
 * $Date: 2001/07/11 16:35:50 $
 */

#include "MarkPlayer.h"
#include <algorithm>
#include "Display.h"

RS_BEGIN_NAMESPACE

// --- CONSTRUCTORS ---

/**
 * Constructor.
 */

MarkPlayer::MarkPlayer()
  : Skill(),
    worldStateView(NULL),
    passTreeView(NULL),
    comView(NULL)
{
  getView(worldStateView, WORLD_STATE_VIEW_ID);
  getView(passTreeView, PASS_TREE_VIEW_ID);
  getView(comView, COMMUNICATION_VIEW_ID);

  // set defaults
  setBuffer();
  setDistance();
  setThreshold();
  setMaxPlayers();
}

// --- DESTRUCTOR ---

MarkPlayer::~MarkPlayer()
{
  viewManager->releaseView(WORLD_STATE_VIEW_ID);
  viewManager->releaseView(PASS_TREE_VIEW_ID);
  viewManager->releaseView(COMMUNICATION_VIEW_ID);
}

// --- METHODS ---

const PlayerState& MarkPlayer::getPlayer() const {
  return target;
}

void MarkPlayer::setMaxPlayers(const unsigned int max) {
  maxPlayers = max;
}

void MarkPlayer::setThreshold(const double min) {
  scoreThreshold = min;
}


void MarkPlayer::setBuffer(const double bufSize) {
  markSkill.setBuffer(distanceBuffer = bufSize);
}

void MarkPlayer::setDistance(const double dist) {
  markSkill.setDistance(distance = dist);
}

bool MarkPlayer::isPlayerMarkingPlayer(const PlayerState& plr,
				       const PlayerState& mark) {
  const double bufferFactor = 1.5;

  double d = (plr.getPosition() - mark.getPosition()).getLength();
  return ( distance - distanceBuffer * bufferFactor < d &&
	   d < distance + distanceBuffer * bufferFactor);
}


bool MarkPlayer::isOpponentMarked(const PlayerState& op) {

  if(op.getUniformNumber() > 0) { // is number known?
    // check with requests 
    std::vector<MarkRequest>::iterator req;
    for(req = requests.begin(); req != requests.end(); req++) 
      if(req->targetNo == op.getUniformNumber())
	return true;
  }
  // Check all teammates against this player
  const std::vector<PlayerState>& tms(worldStateView->getWorldState().getTeammates());
  std::vector<const PlayerState>::iterator tm;
  for(tm = tms.begin(); tm != tms.end(); tm++) {
    if(tm->getUniformNumber() != 1  // Goalie doesn't mark players
       && isPlayerMarkingPlayer(*tm, op))
      return true;
  }
  return false;
}

bool MarkPlayer::isTeammateBusy(const PlayerState& tm) {
  if(tm.getUniformNumber() == 1)
    return true; // goalie is always busy
  
  if(tm.getUniformNumber() > 0) { // is number known?
    // check with requests 
    std::vector<MarkRequest>::iterator req;
    for(req = requests.begin(); req != requests.end(); req++) 
      if(req->playerNo == tm.getUniformNumber())
	return true;
  }
  // Check all opponents against this player
  const std::vector<PlayerState>& ops(worldStateView->getWorldState().getOpponents());
  std::vector<const PlayerState>::iterator op;
  for(op = ops.begin(); op != ops.end(); op++) {
    if(isPlayerMarkingPlayer(tm, *op))
      return true;
  }
  return false;
}

bool MarkPlayer::findTarget() {
  static int lastCallCycle = 0;
  static bool lastReturnValue = false;
  static int lastSuccessCycle = -5;
  const int stickyCycles = 3;

  int currentTime = worldStateView->getWorldState().getGameTime();
  if(lastCallCycle == currentTime)
    return lastReturnValue || (lastSuccessCycle + stickyCycles >= currentTime);
  lastCallCycle = worldStateView->getWorldState().getGameTime();
  lastReturnValue = false;

  PassTree theirPassTree(passTreeView->getOpponentPassTreeWithoutAgent());
  const PassTree& theirPassTreeWithAgent(passTreeView->getOpponentPassTree());

  const std::vector<PlayerState>& tms(theirPassTree.getBadGuys());
  const std::vector<PlayerState>& ops(theirPassTree.getGoodGuys());
  const BallState& ball(worldStateView->getWorldState().getBall());
  const AgentState& agent(worldStateView->getWorldState().getAgent());
  const std::vector<float>& scoreValues(theirPassTree.getScoreValues());
  float ourAdvanceLine = passTreeView->getAgentPassTree().getOurAdvanceLine();
  
  std::vector<unsigned int> candidates;
  for(unsigned int i=0; i < ops.size(); i++) {
    if(scoreValues[i] < scoreThreshold)
      continue;

    if(ops[i].getPosition().getX() > ourAdvanceLine - 5.0)
      continue;

    candidates.push_back(i);

  }

  if(candidates.size() == 0) 
    return false || (lastSuccessCycle + stickyCycles >= currentTime);

  // Sort candidates in descending order
  IndexPredicate pred(scoreValues);
  sort(candidates.begin(), candidates.end(), pred);

//    cout << "Sorted values:" << endl;
//    for(unsigned int i=0; i < candidates.size(); i++) {
//      cout << "val[" << candidates[i] << "]=" << scoreValues[candidates[i]] << endl;
//    }

  // check all candidates
  double distance, myDistance;
  bool iAmClosest=false;
  unsigned int index=0;
  std::vector<const PlayerState>::iterator tm;
  for(unsigned int i=0; i < min(candidates.size(),maxPlayers); i++) {
    if(isOpponentMarked(ops[candidates[i]]))
      continue;
    
    //Check if a non-busy teammate is closer than me
    myDistance = (ops[i].getPosition() - agent.getPosition()).getLength();
    iAmClosest = true;
    index = i;
    for(tm = tms.begin(); tm != tms.end(); tm++) {
      distance = (ops[i].getPosition() - tm->getPosition()).getLength();
      if(distance < myDistance + 0.5 &&
	 !isTeammateBusy(*tm)) {
	iAmClosest = false;
	break;
      }
    }
    if(iAmClosest) 
      break;
  }

  if(iAmClosest) {
    target = ops[index];
//      markSkill.setPosition(TVector(-3, 0) + target.getPosition());
    markSkill.setPosition(target.getPosition());
    theirPassTree.recompute(index);
    if(target.getPosition().getX() < -30) {
      markSkill.setDistantPosition(TVector(-52,5,0));
      markSkill.setLookAtMark(false);
    } else if(theirPassTree.hasParent()) {
      markSkill.setDistantPosition(theirPassTree.getParent().getPosition());
      markSkill.setLookAtMark(true);
      Display::instance()->addArrow(target.getPosition(),
				    theirPassTree.getParent().getPosition(),
				    1, Display::cBlue);
    } else {
      markSkill.setBallAsDistantPosition(true);
      markSkill.setLookAtMark(true);
      Display::instance()->addArrow(target.getPosition(),
				    ball.getPosition(),
				    1, Display::cBlue);
    }
    lastReturnValue = true;
    lastSuccessCycle = currentTime;
    return true;
  }
  return false || (lastSuccessCycle + stickyCycles >= currentTime);
}

Plan MarkPlayer::generatePlan() {
  if(findTarget())
    return markSkill.generatePlan();

  return Plan();
}

FuzzyBool MarkPlayer::persistent() {
  return false;
}

FuzzyBool MarkPlayer::applicable() {
  if(findTarget() && markSkill.applicable())
    return true;
  
  return false;
}

// --- Class MarkRequest ---
MarkPlayer::MarkRequest::MarkRequest() 
  : requestCycle(0),
    targetNo(0),
    playerNo(0)
{
}

MarkPlayer::MarkRequest::~MarkRequest() {
}

bool MarkPlayer::MarkRequest::isValid(const int currentCycle) {
  const int timeToLive = 23;
  return currentCycle - requestCycle < timeToLive;
}

// --- Class IndexPredicate ---
MarkPlayer::IndexPredicate::IndexPredicate(const std::vector<float>& _values) 
  : values(_values)
{
}
 
bool MarkPlayer::IndexPredicate::operator()(const unsigned int left, 
					    const unsigned int right) {
  return values[left] > values[right];
}

RS_END_NAMESPACE







