#include "PointMath.h"
#include <stdio.h>

RS_BEGIN_NAMESPACE

/****************************************
 * Operation for binary Point addition  *
 ****************************************/

std::vector<opType> OpPointPlus::getArgumentTypes()
{
  std::vector<opType> aTypes;
  aTypes.insert(aTypes.end(), opPoint);
  aTypes.insert(aTypes.end(), opPoint);
  return aTypes;
}

Point OpPointPlus::getPoint()
{
  if(arguments.size() >= 2)
    return arguments[0]->getPoint() + arguments[1]->getPoint();
  return Point(0,0);
}

std::string OpPointPlus::getOperationString()
{
  return "p+";
}

/*******************************************
 * Operation for binary Point subtraction  *
 *******************************************/

std::vector<opType> OpPointMinus::getArgumentTypes()
{
  std::vector<opType> aTypes;
  aTypes.insert(aTypes.end(), opPoint); 
  aTypes.insert(aTypes.end(), opPoint);
  return aTypes;
}

Point OpPointMinus::getPoint()
{
  if(arguments.size() >= 2)
    return arguments[0]->getPoint() - arguments[1]->getPoint();
  return Point(0,0);
}

std::string OpPointMinus::getOperationString()
{
  return "p-";
}

/**********************************************
 * Operation for binary Point multiplication  *
 **********************************************/

std::vector<opType> OpPointMultiply::getArgumentTypes()
{
  std::vector<opType> aTypes;
  aTypes.insert(aTypes.end(), opPoint);
  aTypes.insert(aTypes.end(), opPoint);
  return aTypes;
}

float OpPointMultiply::getFloat()
{
  if(arguments.size() >= 2)
    return 
      arguments[0]->getPoint().getX() * arguments[1]->getPoint().getX() + 
      arguments[0]->getPoint().getY() * arguments[1]->getPoint().getY();
  return 0.0;
}

std::string OpPointMultiply::getOperationString()
{
  return "p*";
}

/**********************************************
 * Operation for binary Point scaling         *
 * Multiplication with a float                *
 **********************************************/

std::vector<opType> OpPointScale::getArgumentTypes()
{
  std::vector<opType> aTypes;
  aTypes.insert(aTypes.end(), opFloat);
  aTypes.insert(aTypes.end(), opPoint);
  return aTypes;
}

Point OpPointScale::getPoint()
{
  if(arguments.size() >= 2)
    return arguments[1]->getPoint() * arguments[0]->getFloat();
  return Point(0,0);
}

std::string OpPointScale::getOperationString()
{
  return "pscale";
}

/****************************************
 * Operation for Point constants        *
 ****************************************/

OpPointConstant::OpPointConstant() {
  OpPointConstant(Point());
}

OpPointConstant::OpPointConstant(Point val) {
  value = val;
}


std::vector<opType> OpPointConstant::getArgumentTypes()
{
  std::vector<opType> aTypes;
  aTypes.clear();
  return aTypes;
}

Point OpPointConstant::getPoint()
{
  return value;
}

std::string OpPointConstant::getOperationString()
{
  char buf[256];
  std::snprintf(buf,256, "p x: %f, y: %f", value.getX(), value.getY());
  return buf;
}

/**********************************************
 * Operation to get the length of a "point"   *
 **********************************************/
std::vector<opType> OpPointLength::getArgumentTypes()
{
  std::vector<opType> aTypes;
  aTypes.insert(aTypes.end(), opPoint);
  return aTypes;
}

float OpPointLength::getFloat()
{
  if(arguments.size() >= 1)
    return Vector(arguments[0]->getPoint()).getLength();
  return 0.0;
}

std::string OpPointLength::getOperationString()
{
  return "plength";
}

/**********************************************
 * Operation to get the angle of a "point"    *
 **********************************************/
std::vector<opType> OpPointAngle::getArgumentTypes()
{
  std::vector<opType> aTypes;
  aTypes.insert(aTypes.end(), opPoint);
  return aTypes;
}

float OpPointAngle::getFloat()
{
  if(arguments.size() >= 1)
    return Vector(arguments[0]->getPoint()).getAngle().getDeg();
  return 0.0;
}

std::string OpPointAngle::getOperationString()
{
  return "pangle";
}

/**********************************************
 * Operation to get the X value of a point    *
 **********************************************/
std::vector<opType> OpPointX::getArgumentTypes()
{
  std::vector<opType> aTypes;
  aTypes.insert(aTypes.end(), opPoint);
  return aTypes;
}

float OpPointX::getFloat()
{
  if(arguments.size() >= 1)
    return arguments[0]->getPoint().getX();
  return 0.0;
}

std::string OpPointX::getOperationString()
{
  return "px";
}

/**********************************************
 * Operation to get the Y value of a point    *
 **********************************************/
std::vector<opType> OpPointY::getArgumentTypes()
{
  std::vector<opType> aTypes;
  aTypes.insert(aTypes.end(), opPoint);
  return aTypes;
}

float OpPointY::getFloat()
{
  if(arguments.size() >= 1)
    return arguments[0]->getPoint().getY();
  return 0.0;
}

std::string OpPointY::getOperationString()
{
  return "py";
}

/**********************************************
 * Operation to construct a point from x/y    *
 **********************************************/
std::vector<opType> OpPoint::getArgumentTypes()
{
  std::vector<opType> aTypes;
  aTypes.insert(aTypes.end(), opFloat);
  aTypes.insert(aTypes.end(), opFloat);
  return aTypes;
}

Point OpPoint::getPoint()
{
  if(arguments.size() >= 2)
    return Point(arguments[0]->getFloat(),
		 arguments[1]->getFloat());
  return Point(0,0);
}

std::string OpPoint::getOperationString()
{
  return "point";
}

/**********************************************
 * Operation to construct a point from r/th    *
 **********************************************/
std::vector<opType> OpPolPoint::getArgumentTypes()
{
  std::vector<opType> aTypes;
  aTypes.insert(aTypes.end(), opFloat);
  aTypes.insert(aTypes.end(), opFloat);
  return aTypes;
}

Point OpPolPoint::getPoint()
{
  if(arguments.size() >= 2)
    return Point(Vector(arguments[1]->getFloat(),
			arguments[0]->getFloat()));
  return Point(0,0);
}

std::string OpPolPoint::getOperationString()
{
  return "polpoint";
}

/**********************************************
 * Operation Point arguments                  *
 **********************************************/
std::vector<opType> OpPointArg::getArgumentTypes()
{
  std::vector<opType> aTypes;
  aTypes.insert(aTypes.end(), opFloat);
  return aTypes;
}

Point OpPointArg::getPoint()
{
  int index = 0;
  if (arguments.size() > 0)
    index = floor(arguments[0]->getFloat());
  return getPointArg(index); // getPointArg handles out of bounds cases.
}

std::string OpPointArg::getOperationString()
{
  return "parg";
}

RS_END_NAMESPACE









