/*
 * Ball movement
 * a(t) = Power * Power_rate * n =
 *      = Power * Power_rate * (cos(dir), sin(dir))'
 *
 * dir = dir_kicker + dir_command
 * power_rate = kick_power_rate * (1 - 0.25 * dirDiff / 180 - 0.25 * (dist_ball - player_size - ball_size) / kickable_margin)
 * a = 0.35 = x * kick_power_rate * (1 - 0.25 * dirDiff / 180 - 0.25 * (dist_ball - player_size - ball_size) / kickable_margin)
 * ungefr: x = 0.35 / kick_power_rate
 *
 * Kickable_margin = 0.7
 * kick_power_rate	: 0.016
 * ball_size		: 0.085
 * player_size		: 0.3
 */

/**
 * @file SuperScore.cc
 * Implementation of the SuperScore skill.
 *
 * @par Unit:
 *      Framework/skill
 * $Revision: 1.1.1.1 $
 * $Author: sigge $
 * $Date: 2001/04/08 11:33:58 $
 */

#include "Line.h"
#include "SuperScore.h"
#include "KickCommand.h"
#include "KickBallToSpeedVector.h"
#include "DashToSpeedVectorFacingDirection.h"

#define KICK_RANGE 30
#define SAFE_MARGIN 2
#define CB_FRIENDLY 0.2
#define CB_OPPONENT -1.0
#define CB_ADVANCE  0.5
#define CB_PRECISION 0.0

#define MAXSPEED 2.0
#define EMERGENCY true
RS_BEGIN_NAMESPACE

// --- CONSTRUCTORS ---

/**
 * Constructor.
 */

SuperScore::SuperScore()
  : Skill(),
    parameterView(NULL),
    markerView(NULL),
    agentView(NULL),
    ballView(NULL),
    playerView(NULL)
{
  getView(parameterView, PARAMETER_VIEW_ID);
  getView(playerView, FUZZY_PLAYER_VIEW_ID);
  getView(markerView, MARKER_VIEW_ID, parameterView);
  getView(agentView, AGENT_VIEW_ID, parameterView);
  getView(ballView, FUZZY_BALL_VIEW_ID);
}

// --- DESTRUCTOR ---

SuperScore::~SuperScore()
{
  viewManager->releaseView(PARAMETER_VIEW_ID);
  viewManager->releaseView(MARKER_VIEW_ID);
  viewManager->releaseView(AGENT_VIEW_ID);
  viewManager->releaseView(FUZZY_BALL_VIEW_ID);
  viewManager->releaseView(FUZZY_PLAYER_VIEW_ID);
}


// --- METHODS ---


Vector SuperScore::calcKickAcc(Vector kick, Point bpos, Point agpos, float agdir)
{
/* a(t) = Power * Power_rate * n =
 *      = Power * Power_rate * (cos(dir), sin(dir))'
 *
 * dir = dir_kicker + dir_command
 * power_rate = kick_power_rate * (1 - 0.25 * dirDiff / 180 - 0.25 * (dist_ball - player_size - ball_size) / kickable_margin)
 * a = 0.35 = x * kick_power_rate * (1 - 0.25 * dirDiff / 180 - 0.25 * (dist_ball - player_size - ball_size) / kickable_margin)
 * ungefr: x = 0.35 / kick_power_rate
 *
 * Kickable_margin = 0.7
 * kick_power_rate	: 0.016
 * ball_size		: 0.085
 * player_size		: 0.3
 */

  float dirFactor = 0.25 * abs(Vector(bpos - agpos).getAngle().getDeg() - agdir) / 180.0;
  float distFactor = 0.25 * (Vector(bpos - agpos).getLength()
			     - parameterView->SP_player_size
			     - parameterView->SP_ball_size) / parameterView->SP_kickable_margin;
  float power_rate = parameterView->SP_kick_power_rate * (1 - distFactor - dirFactor);

  return Vector(kick.getAngle(), kick.getLength() * power_rate);
}

Vector SuperScore::calcKickVector(Vector acc, Point bpos, Point agpos, float agdir)
{
/* a(t) = Power * Power_rate * n =
 *      = Power * Power_rate * (cos(dir), sin(dir))'
 *
 * dir = dir_kicker + dir_command
 * power_rate = kick_power_rate * (1 - 0.25 * dirDiff / 180 - 0.25 * (dist_ball - player_size - ball_size) / kickable_margin)
 * a = 0.35 = x * kick_power_rate * (1 - 0.25 * dirDiff / 180 - 0.25 * (dist_ball - player_size - ball_size) / kickable_margin)
 * ungefr: x = 0.35 / kick_power_rate
 *
 * Kickable_margin = 0.7
 * kick_power_rate	: 0.016
 * ball_size		: 0.085
 * player_size		: 0.3
 */

  float dirFactor = 0.25 * abs(Vector(bpos - agpos).getAngle().getDeg() - agdir) / 180.0;
  float distFactor = 0.25 * (Vector(bpos - agpos).getLength()
			     - parameterView->SP_player_size
			     - parameterView->SP_ball_size) / parameterView->SP_kickable_margin;
  float power_rate = parameterView->SP_kick_power_rate * (1 - distFactor - dirFactor);

  return Vector(acc.getAngle(), acc.getLength() / power_rate);
}

Plan SuperScore::genPlan(Point bpos, Vector bspeed, Point agpos, Vector agspeed, float agdir, Plan curPlan)
{
  // Examine our options.
  // a) Kick 100
  // b) Kick x Kick 100
  // c) Turn x, Kick 100

#define destDir 0
  //agdir
#define BODY_DIST 0.5
#define BACK_ANGLE 130
#define FRONT_ANGLE 55

  cout << "================================" << endl;

  Point nextAgpos = agpos + agspeed;
  Vector nextAgspeed = agspeed * parameterView->SP_player_decay;
  Point nextBpos = bpos + bspeed;
  Vector nextBspeed = bspeed * parameterView->SP_ball_decay;

  float bdir = Vector(bpos - agpos).getAngle().getDeg();
  bdir -= destDir;

  Command *cmd;

//    if(abs(bdir) < 90)
//      {

  if(bspeed.getLength() > 0.6
     && abs(bspeed.getAngle().getDeg() - destDir) < 10)
    {
      // The ball is moving roughly in the right direction.
      // Accelerate it.
      cout << "Finalizing!" << endl;
      
      float dirDiff = bspeed.getAngle().getDeg() - destDir;
      Vector tmp = Vector(destDir, bspeed.getLength() * std::cos(M_PI * dirDiff / 180.0));
      Vector comp = bspeed - tmp;
      
      Vector kv = calcKickVector(comp, bpos, agpos, agdir);
      if(kv.getLength() > 99)
	kv.setLength(99);
      
      float len = std::sqrt(pow(100, 2) - pow(kv.getLength(), 2));
      Vector kv2 = kv + Vector(destDir, len);

      cout << "kv2: " << kv2 << endl;
      cout << "Final speed: " << (bspeed + calcKickAcc(kv2, bpos, agpos, agdir)).getLength() << endl;

      cmd = new KickCommand(kv2.getLength(), kv2.getAngle().getDeg() - agdir);
      curPlan.addStep(Step(cmd));
      
      delete cmd;
      return curPlan;
    }
  //    }
  
  // Treat everything relative
  //    x
  //    ^
  //  --|-->y
  //    |
  
  // x points in desiredDir.

  Vector margin = Vector(destDir + 90, BODY_DIST);
  Vector dir = Vector(destDir, 1);
  Line right(Point(nextAgpos + margin), dir); 
  Line left(Point(nextAgpos - margin), dir); 

  Line rback(nextAgpos, Vector(destDir + BACK_ANGLE, 1));
  Line rfront(nextAgpos, Vector(destDir + FRONT_ANGLE, 1));

  Line lback(nextAgpos, Vector(destDir - BACK_ANGLE, 1));
  Line lfront(nextAgpos, Vector(destDir - FRONT_ANGLE, 1));
 
//    Vector margin = Vector(90, BODY_DIST);
//    Vector dir = Vector(0, 1);

//    Line right = Line(Point(margin), dir);
//    Line left = Line(Point(-margin), dir);

//    Line rback(Point(0,0), Vector(BACK_ANGLE, 1));
//    Line rfront(Point(0,0), Vector(FRONT_ANGLE, 1));
  
//    Line lback(Point(0,0), Vector(-BACK_ANGLE, 1));
//    Line lfront(Point(0,0), Vector(-FRONT_ANGLE, 1));

  Point dstPos[2];
  if(bdir > 0)
    { // Right side
      dstPos[0] = right.getIntersection(rfront);
      dstPos[1] = right.getIntersection(rback);

      Vector tmp = Vector(dstPos[0] - nextAgpos);
      tmp.rotate(-destDir);
      cout << "right: " << Point(tmp);
      tmp = Vector(dstPos[1] - nextAgpos);
      tmp.rotate(-destDir);
      cout << " " << Point(tmp) << endl;

      //      cout << "right: " << dstPos[0] << " " << dstPos[1] << endl;
    }
  else
    { // Left side..
      dstPos[0] = left.getIntersection(lfront);
      dstPos[1] = left.getIntersection(lback); 

      Vector tmp = Vector(dstPos[0] - nextAgpos);
      tmp.rotate(-destDir);
      cout << "left: " << Point(tmp);
      tmp = Vector(dstPos[1] - nextAgpos);
      tmp.rotate(-destDir);
      cout << " " << Point(tmp) << endl;

//        cout << "right: " << dstPos[0] << " " << dstPos[1] << endl;
    }

  // We wouldn't be here if we had enough forwardspeed.
  // So if the ball is close to dstPos[0] we would like to have it at dstPos[1].
  // If ball close to dstPos[1]
  //    If can Kickto dstPos[0]
  //        Kickto dstPos[0]
  //    else
  //        Kickto dstPos[1]
  // If ball close to dstPos[0]
  //        Kickto dstPos[1]


  Vector kv;
  Point dbgPos;

  Vector relBspeed = bspeed;
  relBspeed.rotate(-destDir);
  Vector relAgspeed = agspeed;
  relAgspeed.rotate(-destDir);
  Vector relBpos = Vector(bpos);
  relBpos.rotate(-destDir);

  // Check if agent is blocking the cur to front path.

  Line frLine = Line(Point(bpos), Point(dstPos[0]));
  bool blocked = false;
  if(frLine.isThroughCircle(nextAgpos, 1.1 * (parameterView->SP_player_size + parameterView->SP_ball_size)))
    {
      cout << "Frontpath blocked!" << endl;
      blocked = true;
    }
  

  if(abs(bdir) > 90
     && !blocked) // close to dstPos[1]
    {
      //      Vector accNeed = Vector(dstPos[0] + nextAgpos - nextBpos);
      Vector accNeed = Vector(dstPos[0] - nextBpos);
      //      accNeed.rotate(destDir);
      dbgPos = dstPos[0];
      kv = calcKickVector(accNeed, bpos, agpos, agdir);
      if(kv.getLength() > 100)
	{
	  if(bspeed.getLength() > 0.1)
	    {
	      cout << "Stopping ball." << endl;
	      dbgPos = dstPos[1];
	      //	  Vector accNeed = Vector(dstPos[1] + nextAgpos - nextBpos);
	      Vector accNeed = Vector(dstPos[1] - nextBpos);
	      //	  accNeed.rotate(destDir);
	      kv = calcKickVector(accNeed, bpos, agpos, agdir);
	      
	      cout << "Needed: " << accNeed << " kickv: " << kv << endl;
	      if(kv.getLength() > 100)
		kv.setLength(100);
	    }
	  else
	    {
	      cout << "Partial frontkick." << endl;
	      cout << "Needed: " << accNeed << " kickv: " << kv << endl;
	      if(kv.getLength() > 100)
		kv.setLength(100);
	      
	    }
	}
      else
	{
	  cout << "Frontkicking ball." << endl;
	  cout << "Needed: " << accNeed << " kickv: " << kv << endl;
	}
    }
  else  // Close to dstPos[0]
    {
      cout << "Backkicking ball." << endl;

      //      Vector accNeed = Vector(dstPos[1] + nextAgpos - nextBpos);
      Vector accNeed = Vector(dstPos[1] - nextBpos);
      //      accNeed.rotate(destDir);
      dbgPos = dstPos[1];
      kv = calcKickVector(accNeed, bpos, agpos, agdir);
      cout << "Needed: " << accNeed << " kickv: " << kv << endl;
      if(kv.getLength() > 100)
	kv.setLength(100);

    }
  
  Vector relPos = Vector(dbgPos - nextAgpos);
  relPos.rotate(-destDir);
  Vector relPos2 = Vector(bpos - agpos);
  relPos2.rotate(-destDir);
  //  dbgPos.rotate(-destDir);
  Vector kacc = calcKickAcc(kv, bpos, agpos, agdir);
  kacc.rotate(-destDir);
  cout << "curpos: " << Point(relPos2) << "despos: " << Point(relPos) << " kick: " << Point(kacc) << endl;

  cmd = new KickCommand(kv.getLength(), kv.getAngle().getDeg() - agdir);
  curPlan.addStep(Step(cmd));
  delete cmd;

  nextBspeed = nextBspeed + calcKickAcc(kv, bpos, agpos, agdir);
  nextBpos = nextBpos + calcKickAcc(kv, bpos, agpos, agdir);

  return genPlan(nextBpos, nextBspeed, nextAgpos, nextAgspeed, agdir, curPlan);
}

/**
 * Genrates a plan for achieving a pass.
 * Chooses the closest of the passvictims if the argument is
 * supplied, otherwise pass the best player according to 
 * TerritoryControlView
 */

Plan SuperScore::generatePlan()
{

  Point ballPos;
  Vector ballSpeed;
  Point agentPos;
  Vector agentSpeed;
  float dir;

  ballPos = Point(ballView->getAbsPosition(NOW).getVector());
  ballSpeed = ballView->getAbsSpeed(NOW).getVector();

  agentPos = agentView->getAgentPosition(NOW).getPoint();
  agentSpeed = agentView->getSpeedVector(NOW).getVector();
  dir = agentView->getBodyDirection(NOW).getAngle().getAngle().getAngle().getDeg();

  Plan p;
  p = genPlan(ballPos, ballSpeed, agentPos, agentSpeed, dir, p);

  return p;
}

/**
 * Persistent.
 * FALSE
 */
FuzzyBool SuperScore::persistent()
{
  return FALSE;
}

/**
 * Returns the applicable value
 */

FuzzyBool SuperScore::applicable()
{
  
  // Need to know where the ball is.

//    cout << "A" << flush;

  if(!ballView->getAbsPosition().isKnown())
    return FALSE;

//    cout << "B" << flush;

  if(ballView->getAbsPosition(NOW).getLength().getConfidence() < 0.1)
    return FALSE;

//    cout << "C" << flush;
  // We need to be able to kick the ball.
  if(ballView->getRelPosition(NOW).getLength() < parameterView->SP_kickable_area)
    return TRUE;

//    cout << "D" << flush;
  return FALSE;
}

RS_END_NAMESPACE

