#include "AngleThingy.h"
#include "PlayerState.h"
#include "ViewManager.h"

RS_BEGIN_NAMESPACE

// #define MATLAB_DUMP

/** Constructor.
 */
AngleThingy::AngleThingy() :
  worldView(NULL),
  parameterView(NULL)
{
#ifdef MATLAB_DUMP
  matlab = fopen("test.m", "wt");
#endif

  getView(worldView, WORLD_STATE_VIEW_ID);
  getView(parameterView, PARAMETER_VIEW_ID);
  reset();
}
/** Destructor.
 */
AngleThingy::~AngleThingy()
{
  ViewManager* viewManager = ViewManager::instance();
  viewManager->releaseView(WORLD_STATE_VIEW_ID);
  viewManager->releaseView(PARAMETER_VIEW_ID);

#ifdef MATLAB_DUMP
  fclose(matlab);
#endif
}

/** Resets the anglethingy to start, clears spans, resets gametime and
 * initalizes to the current worldstate.
 */
void AngleThingy::reset()
{
  spans.clear();
  time = 0;
  worldState = worldView->getWorldState();
  ball = worldState.getBall();
}

/** Adds a span, currently only works with empty AngleThingys. Use this
 * method to set the initial span that the anglethingy should work with.
 */
void AngleThingy::addSpan(const AngleThingy::Span& span)
{
  spans.push_back(span);
}

/** Sets the worldstate that the worldstate should work upon. If this method
 * isn't called anglethingy will work on the worldstate that was current
 * when the AngleThingy object was created.
 */
void AngleThingy::setWorldState(const WorldState& ws)
{
  worldState = ws;
  ball = worldState.getBall();
}

/** Private help method.
 */
void AngleThingy::removeSpan(const AngleThingy::Span& s)
{
  list<AngleThingy::Span>::iterator spanIt;
  for(spanIt = spans.begin(); spanIt != spans.end(); spanIt++)
  {
    // Check if we're done.
    if(s.end < spanIt->start)
    {
      break; // We're done.
    }
    else
    {
      if(s.start < spanIt->end)
      {
	if(s.start < spanIt->start
	   && s.end > spanIt->end)
	{
	  // Remove the old span.
	  //  	      cout << "Remove span." << endl;
	  spanIt = spans.erase(spanIt);
	}
	else if(s.start < spanIt->start)
	{
	  // Cut of start piece.
	  //  	      cout << "Trim start." << endl;
	  spanIt->start = s.end;
	}
	else if(s.end > spanIt->end)
	{
	  // Cut of end piece;
	  //  	      cout << "Trim end." << endl;
	  spanIt->end = s.start;
	}
	else
	{
	  // Split span into subspans.
	  //  	      cout << "Split span." << endl;
	  AngleThingy::Span newSpan;
	  newSpan.start = spanIt->start;
	  newSpan.end = s.start;
	  spanIt->start = s.end;
	  spanIt = spans.insert(spanIt, newSpan);
	}
      }
    }
  }
}

/** Sets the power for the kick. This is used to calculate the radial 
 * positions of the ball. Anglethingy will compensate for slides when 
 * using the method.
 * If a special world state is needed it should be set prior to calling
 * this method.
 */
void AngleThingy::setKickPower(float p)
{
  compensateSlide = true;

  // Assume max kick.
  double kpr = 
    worldState.getAgent().getKickPowerRate(worldState.getBall());
  v0 = kpr * 100.0;
}

/** Sets the initial acceleration for the kick. This is used to calculate
 * the radial positions of the ball. Anglethingy will compensate for slides
 * when using the method.
 */
void AngleThingy::setKickAcceleration(float a)
{
  compensateSlide = true;
  v0 = a;
}

/** Sets the initial speed used for the calculations, using this method
 * will result in no compensation for slides.
 */
void AngleThingy::setInitialSpeed(float v)
{
  compensateSlide = false;
  v0 = v;
}

/** Gets a list of spans that are safe after the number of steps that has been
 * performed.
 */
const list<AngleThingy::Span> AngleThingy::getSpans()
{
  return spans;
}

/** Returns the action radius the ball has for the current step.
 */
float AngleThingy::getBallRadius()
{
  return ballRadius;
}

/** Returns the ballstate that is associated with the current step.
 */
const BallState& AngleThingy::getBall()
{
  return ball;
}

/** Help function that dumps a circle to a matlab debug file.
 */

void AngleThingy::dumpCircle(TVector p, double r)
{
  fprintf(matlab, "cir = [");
  for(int i = -180; i <= 180; i+=10)
  {
    double x = r*std::cos((float)i*M_PI/180.0) + p.getX();
    double y = r*std::sin((float)i*M_PI/180.0) + p.getY();
    fprintf(matlab, "(%f+i*%f) ", x, y);
  }
  fprintf(matlab, "];plot(cir, 'r');\n");
}

/** Help function that dumps the spans to a matlab debug file.
 */
void AngleThingy::dumpSpans()
{
  if(spans.size() == 0)
    return;

  fprintf(matlab, "spans = [");
  list<AngleThingy::Span>::const_iterator itSpans = spans.begin();
  for(int i = -180; i < 180 && itSpans != spans.end(); i+=1)
  {
    while(itSpans != spans.end() && i > itSpans->end)
      itSpans++;
    
    if(i > itSpans->start && i < itSpans->end && itSpans != spans.end())
    {
      fprintf(matlab, "(%f+i*%f) ", 
	      ballRadius * std::cos(M_PI*(float)i/180.0),
	      ballRadius * std::sin(M_PI*(float)i/180.0));
    }
    else
    {
      fprintf(matlab, "(0+i*eps) "); 
    }
  }
  fprintf(matlab, "];plot(spans);hold on;\n");
}

/** Help function that dumps opponents to matlab debugfile.
 */
void AngleThingy::dumpOpponents()
{
  // MATLAB
  const vector<PlayerState>& opponents = worldState.getOpponents();
  vector<PlayerState>::const_iterator itOpponents;

  if(opponents.size() > 0)
  {
    fprintf(matlab, "opps = [");
    for(itOpponents = opponents.begin();
	itOpponents != opponents.end();
	itOpponents++)
    {
      fprintf(matlab, "(%f+i*%f) ", 
	      (itOpponents->getPosition()-ball.getPosition()).getX(),
	      (itOpponents->getPosition()-ball.getPosition()).getY());
    }
    fprintf(matlab, "];plot(opps, 'xr');");

    for(itOpponents = opponents.begin();
	itOpponents != opponents.end();
	itOpponents++)
    {
      dumpCircle(itOpponents->getPosition() - ball.getPosition(), 
		 getActionRadius());
    }

  }
}

void AngleThingy::dumpAgent()
{
  // MATLAB
  fprintf(matlab, "plot((%f+i*%f),'og');hold on;\n ", 
	  (worldState.getAgent().getPosition()-ball.getPosition()).getX(),
	  (worldState.getAgent().getPosition()-ball.getPosition()).getY());
}

float AngleThingy::getActionRadius()
{
  // Calculate opponent action radius.
  float actionRadius = 0.0;
  if(time > actionPenalty)
    actionRadius = (time-actionPenalty) 
      * opponentEffect * parameterView->SP_player_speed_max;
  if(time >= observationPenalty)
    actionRadius += 1.5 * parameterView->SP_kickable_area;
  else
    actionRadius += 1.5 * parameterView->SP_player_size;
  return actionRadius;
}

AngleThingy::Span AngleThingy::getMaxSpan()
{
  AngleThingy::Span maxSpan;
  float maxSpanSize = -1;
  //      cout << "{";
  list<AngleThingy::Span>::const_iterator itSpans;
  for(itSpans = spans.begin(); itSpans != spans.end(); itSpans++)
  {
    if( (itSpans->end - itSpans->start) > maxSpanSize )
    {
      maxSpanSize = itSpans->end - itSpans->start;
      maxSpan = *itSpans;
    }
  }
  return maxSpan;
}

/** Steps the anglethingy and trims away unsafe directions. Unsafe is defined
 * as the directions where at the current time the ball can be intercepted
 * by an opponent.
 */
void AngleThingy::step()
{
  // Get the vector with opponents.
  const vector<PlayerState>& opponents = worldState.getOpponents();
  vector<PlayerState>::const_iterator itOpponents;
  
  // Calculate the ball action radius.
  // FIXME: This should consider how hard we can kick the ball
  ballRadius = 
    v0 * (pow(parameterView->SP_ball_decay, time) - 1) 
    / (parameterView->SP_ball_decay - 1);
  
  for(itOpponents = opponents.begin();
      itOpponents != opponents.end();
      itOpponents++)
  {
    // Calculate delta.
    TVector delta = itOpponents->getPosition() - ball.getPosition();
    float d = delta.getLength();

    float actionRadius = getActionRadius();

    if( (d + ballRadius) < actionRadius)
    {
      // Player's action radius completely covers the ball radius.

      // No directions are safe anymore.
      spans.clear();
    }
    else if( (d + actionRadius) < ballRadius )
    {
      // The ball has passed the player and he cannot reach it (for a while).
      // (BallRadius completely covers his action radius

      // Do nothing.
    }
    else if( d > (actionRadius + ballRadius) )
    {
      // The ball and and the action circles doesn't intersect.
      
      // Do nothing.
    }
    else
    {
      // The circles intersects.

      float a = 
	(ballRadius*ballRadius - actionRadius*actionRadius + d*d) / (2 * d);

      float h2 = ballRadius*ballRadius - a*a;
      if( h2 < 0 )
      {
	cout << "Something is teribly wrong... h^2 < 0" << endl;
      }

      float h = std::sqrt(h2);
      TVector p2 = a * delta / d;
      
      TVector r = h * TVector(delta.getY(), -delta.getX()) / d;

      TVector p3a = p2 + r;
      TVector p3b = p2 - r;

      // Calculate the span that he covers.
      AngleThingy::Span s;
      s.start = p3a.getAngle();
      s.end = p3b.getAngle();

      if(s.end < s.start)
      {
	AngleThingy::Span s1;
	s1.start = -179.9;
	s1.end = s.end;
	removeSpan(s1);
	s1.start = s.start;
	s1.end = 179.9;
	removeSpan(s1);
      }
      else
      {
	removeSpan(s);
      }

      if(spans.size() == 0)
	break;
    }
  }

#ifdef MATLAB_DUMP
  dumpSpans();
  dumpAgent();
  dumpOpponents();
#endif

  time++;
  if(compensateSlide)
    ball.step();
  ball.setC(1.0);
}

RS_END_NAMESPACE
