/*
 * $Id: BallState.cc,v 1.18 2001/07/30 16:51:41 andbo928 Exp $
 * 
 * HISTORY 
 *
 * $Log: BallState.cc,v $
 * Revision 1.18  2001/07/30 16:51:41  andbo928
 * Multistep now updates gametime.
 *
 * Revision 1.17  2001/07/30 15:22:23  andbo928
 * Fixed a minor bug.
 *
 * Revision 1.16  2001/07/29 17:08:31  mikbr212
 * Added step(n)
 *
 * Revision 1.15  2001/07/28 14:08:12  andbo928
 * Fixed a minor bug in step().
 *
 * Revision 1.14  2001/07/01 12:37:56  andbo928
 * Added new limitations imposed by the new server in simulation.
 *
 * Revision 1.13  2001/06/05 20:33:53  mikbr212
 * Fixed doxygen-stubs for all classes
 *
 */

/**
 * @file BallState.cc
 * Implementation file for the class BallState.
 *
 * @par Unit:
 *      Framework/view
 * $Revision: 1.18 $
 * $Author: andbo928 $
 * $Date: 2001/07/30 16:51:41 $
 */

#include "BallState.h"
#include "ViewPack.h"

RS_BEGIN_NAMESPACE

extern ViewPack* viewPack;
//ParameterView* BallState::parameterView;

BallState::BallState() :
  speed(TVector(0,0)),
  position(TVector(0,0)),
  kickAcceleration(TVector(0,0)),
  gameTime(0)
{
  if(viewPack == NULL)
    {
      viewPack = new ViewPack();
    }
  parameterView = viewPack->parameterView;
}

bool BallState::operator==(const BallState& other) const
{
  return 
    speed == other.speed
    && position == other.position
    && gameTime == other.gameTime;
}

/** Steps the ball n cycles. Cheaper than calling step n times.
 */
void BallState::step(int cycles) {
  if(cycles < 1)
    return;
  step(); // fix acceleration and stuff like that
  cycles--;

  if(cycles < 1)
    return;
  float d = parameterView->SP_ball_decay;
  float dPowN = pow(d, cycles - 1);
  position += (1 - dPowN) / (1 - d) * speed;
  speed *= dPowN;

  setGameTime(getGameTime() + cycles - 1);
  
  setC(getC() * pow(0.9, cycles));
}

void BallState::step()
{
  if(kickAcceleration.getLength() > parameterView->SP_ball_accel_max)
    kickAcceleration.setLength(parameterView->SP_ball_accel_max);

  TVector u = speed + kickAcceleration;

  if(u.getLength() > parameterView->SP_ball_speed_max)
    u.setLength(parameterView->SP_ball_speed_max);

  if(u.getC() > 0.5)
    position = position + u;

  speed = u * parameterView->SP_ball_decay;

  kickAcceleration = TVector(0,0,1);

  setC(getC() * 0.9);

  gameTime++;
}

void BallState::init(const BallObject& ballObject)
{
  setPosition( TVector(ballObject.getPosition()) );
  setSpeed( TVector(ballObject.getAbsoluteSpeedVector()) );
}

bool BallState::isHeadedTo(const TVector& pos, 
			   const double margin,
			   const unsigned int cycles) const {
  if(getC() < 0.3)
    return false;
  
  if(speed.getLength() < 0.1) 
    return false;

  double scale;
  if(cycles > 0)
    scale = (1 - pow(parameterView->SP_ball_decay,cycles)) / 
      (1-parameterView->SP_ball_decay);
  else
    scale = 1.0 / (1-parameterView->SP_ball_decay);
  
  TLine ballLine(position, position + speed*scale);
  return ballLine.isThroughCircle(pos,margin);
}

void BallState::setAcceleration(const TVector& accVec)
{
  kickAcceleration = accVec;
}

const TVector& BallState::getAcceleration() const
{
  return kickAcceleration;
}

//  void BallState::doKick(TVector kickVector, const PlayerState& kicker)
//  {
//    TVector kToB = position - kicker.getPosition();

//    // FIXME: Modulo.
//    float dirFactor = 
//      0.25 
//      * abs(kToB.getAngle()
//  	  - kicker.getDirection()) / 180.0;

//    float distFactor = 
//      0.25 
//      * (kToB.getLength()
//         - parameterView->SP_player_size
//         - parameterView->SP_ball_size) 
//      / parameterView->SP_kickable_margin;

//    float power_rate = 0;
//    if(kToB.getLength() < parameterView->SP_kickable_area)
//      {
//        power_rate = 
//  	parameterView->SP_kick_power_rate 
//  	* (1 - distFactor - dirFactor);
//      }

//    TVector a = 
//      TVector::TPolarVector(kickVector.getLength() * power_rate,
//  			  (kickVector.getAngle() 
//  			   +kicker.getDirection()));
//    kickAccel = kickAccel + a;
//  }

//  void BallState::doNull()
//  {
//    // Intentionally left blank.
//  }

TVector BallState::getSpeed() const
{
  return speed;
}

void BallState::setSpeed(const TVector &value)
{
  speed = value;
}

TVector BallState::getPosition() const
{
  return position;
}

void BallState::setPosition(const TVector &value)
{
  position = value;
}

int BallState::getGameTime() const
{ 
  return gameTime;
}

void BallState::setGameTime(int _gameTime)
{ 
  gameTime = _gameTime;
}

double BallState::getC() const {
  return position.getC();
  //	     speed.getC());
}

void BallState::setC(double _c) {
  position.setC(_c);
  speed.setC(_c);
}

string BallState::serialize() const
{

  string result; short Short; char Char; int Int; float Float;

  result += asString(Short = (32000.0 * position.getX() / 60.0));
  result += asString(Short = (32000.0 * position.getY() / 40.0));
  result += asString(Short = (32000.0 * position.getC()));

  result += asString(Short = (32000.0 * speed.getX() / 4.0));
  result += asString(Short = (32000.0 * speed.getY() / 4.0));
  result += asString(Short = (32000.0 * speed.getC()));

  return result;
}

string BallState::deserialize(const string& buf)
{

  string result = buf; short Short; char Char; int Int;
  float _x, _y, _c;

  _x = fromString(Short, result) * 60.0 / 32000.0;
  result = skipString(Short, result);
  _y = fromString(Short, result) * 40.0 / 32000.0;
  result = skipString(Short, result);
  _c = fromString(Short, result) / 32000.0;
  result = skipString(Short, result);
  position = TVector(_x, _y, _c);

  _x = fromString(Short, result) * 4.0 / 32000.0;
  result = skipString(Short, result);
  _y = fromString(Short, result) * 4.0 / 32000.0;
  result = skipString(Short, result);
  _c = fromString(Short, result) / 32000.0;
  result = skipString(Short, result);
  speed = TVector(_x, _y, _c);

  return result;
}

RS_END_NAMESPACE
