/* -*- Mode: C++ -*-
 * $Id: LookView.cc,v 1.11 2001/07/28 07:18:05 mikbr212 Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: LookView.cc
 * Unitname: Framework/view
 * $Revision: 1.11 $
 * Created by: Fredrik Heintz 2000-09-12
 * Last modified by $Author: mikbr212 $ $Date: 2001/07/28 07:18:05 $
 *
 *
 * HISTORY
 *
 * $Log: LookView.cc,v $
 * Revision 1.11  2001/07/28 07:18:05  mikbr212
 * Removed warnings!!
 *
 * Revision 1.10  2001/07/26 15:44:59  mikbr212
 * Computes values on a radius reduced world state now
 *
 * Revision 1.9  2001/07/26 12:51:24  mikbr212
 * Adjusted constant
 *
 * Revision 1.8  2001/07/06 10:00:38  mikbr212
 * Merged updates and resolved conflicts (a lot)
 *
 * Revision 1.7  2001/04/23 15:05:34  sigge
 * Goalie now uses narrow vision
 *
 * Revision 1.6  2001/04/16 14:51:38  mikbr212
 * removed cout
 *
 * Revision 1.5  2001/04/16 13:31:27  mikbr212
 * Debugged view
 *
 * Revision 1.4  2001/04/13 21:45:49  andbo928
 * It now compiles on my machine. Grrr. Ugly initializer code thou.
 *
 * Revision 1.3  2001/04/13 14:55:39  mikbr212
 * added stuff
 *
 * Revision 1.2  2001/04/13 10:23:46  mikbr212
 * Now using worldStateView
 *
 * Revision 1.1.1.1  2001/04/08 11:33:59  sigge
 * Imported sources
 *
 * Revision 1.8  2000/10/28 20:22:02  mikbr212
 * changed initBall: now ballSpeed is retrieved from commandView if not known in ballView
 *
 * Revision 1.7  2000/10/26 16:00:45  mikbr212
 * Fixed bug in kickPowerRate: kickable_area to kickable_margin
 *
 * Revision 1.6  2000/10/09 18:46:15  mikbr212
 * Fixed serious wrong-calculations in getKickVector
 *
 * Revision 1.5  2000/10/09 10:17:44  mikbr212
 * added comments
 *
 * Revision 1.4  2000/10/08 20:05:01  mikbr212
 * Removed spam messages
 *
 * Revision 1.3  2000/10/08 12:57:28  mikbr212
 * removed comment
 *
 * Revision 1.2  2000/10/07 19:42:01  olalu526
 * Uncommented part using line.
 *
 * Revision 1.1  2000/10/07 14:29:25  mikbr212
 * created
 *
 */

/**
 * @file LookView.cc
 * Implementation file for the view class LookView.
 *
 * @par Unit:
 *      Framework/view
 * $Revision: 1.11 $
 * $Author: mikbr212 $
 * $Date: 2001/07/28 07:18:05 $
 */


#include "Display.h"
#include "LookView.h"

RS_BEGIN_NAMESPACE

/**
 * Constructor. Use getView instead.
 * @see getView
 */
LookView::LookView(const ViewId id, int _slices)
  : View(id),
    slices(_slices),
    see(slices,0),
    lookValues(slices,0),
    integralLookValues(slices,0),
    worldStateView(NULL),
    parameterView(NULL)
{
  getView(worldStateView, WORLD_STATE_VIEW_ID);
  getView(parameterView, PARAMETER_VIEW_ID);

  setWeights();
}

LookView::~LookView()
{
  viewManager->releaseView(WORLD_STATE_VIEW_ID);
  viewManager->releaseView(PARAMETER_VIEW_ID);
}

void LookView::setWeights(Float ws, Float wp, Float wb) {
  Float sum = ws + wp + wb;
  wSee = ws / sum;
  wPlr = wp / sum;
  wBall = wb / sum;

  computeLookValues();
}

Float LookView::getSlice(AngleDeg a) {
  Float retVal = a.getDeg() * Float(slices) / 360.0;
  while(retVal < 0)
    retVal += slices;
  while(retVal >= slices)
    retVal -= slices;
  return retVal;
}

void LookView::computeLookValues() {
  WorldState ws = worldStateView->getWorldState();
  ws.reduceR((parameterView->SP_unum_far_length +
	      parameterView->SP_unum_far_length)/2);
  
  TVector agentPos = ws.getAgent().getPosition();

  // compute ball-slice
  int ballIndex = -1;
  Float ballF = 0;
  if(ws.getBall().getPosition().getC() > 0.01) {
    ballIndex = getIndex((ws.getBall().getPosition() - agentPos).getAngle());
    ballF = 1 - ws.getBall().getPosition().getC();
  }

  // Clear old values and copy see-values
  {
    std::vector<Float>::iterator it;
    std::vector<Float>::iterator seeIt;

    // if ball was not seen, add wBall/slices to all slices
    Float addToAll = 0;
    if(ballIndex == -1)
      addToAll += wBall/slices;

    seeIt = see.begin();
    for(it = lookValues.begin(); it != lookValues.end() && seeIt != see.end(); it++) {
      *it = (1-*seeIt) * wSee + addToAll;
      seeIt++;
    }
  }

  // Add ball if known
  if(ballIndex != -1)
    lookValues[ballIndex] += ballF * wBall;

  // compute plr-slices
  std::vector<PlayerState> op = ws.getOpponents(); 
  std::vector<PlayerState> tm = ws.getTeammates();
  Float max_players_per_slice = Float(max(int(20),
					  int(tm.size()+op.size()))) / slices;

  std::vector<PlayerState>::iterator it;
  // Opponents
  for(it = op.begin(); it != op.end(); it++) {
    lookValues[getIndex((it->getPosition() - agentPos).getAngle())] += 
      (1 - it->getPosition().getC()) / max_players_per_slice * wPlr;
  }
  // Teammates
  for(it = tm.begin(); it != tm.end(); it++) {
    lookValues[getIndex((it->getPosition() - agentPos).getAngle())] += 
      (1 - it->getPosition().getC()) / max_players_per_slice * wPlr;
  }
  
  // ** Compute integralLookValues as sum of the view cone **

  {
    int offset, cone(slices/4);
    if(sensorInterface)
    {
      switch (sensorInterface->getViewWidth())
      {
      case VW_UNKNOWN:
      case VW_NORMAL:
	cone = slices / 4; // 90 degrees
	break;
      case VW_WIDE:
	cone = slices / 2; // 180 degrees
	break;
      case VW_NARROW:
	cone = slices / 8; // 45 degrees
	break;
      }
    }

    offset = cone/2;

    // compute first value
    std::vector<Float>::iterator it;
    it = integralLookValues.begin();
    *it = 0;
    for(int i=slices; i < cone+slices; i++) {
      *it += lookValues[(i-offset)%slices] / (Float) cone;
    }

    // Compute the rest
    for(int i=slices+1; i<slices+slices; i++) {
      integralLookValues[i%slices] = integralLookValues[(i-1)%slices] -
	lookValues[(i-offset)%slices]/(Float)cone + lookValues[(i+offset-1)%slices]/(Float)cone;
    }
  }      
}

/** Returnes the maximum value by searching in all directions. This gives
 * a kind of measure of the need of looking around.
 * A high value means low confidence in the things in the world.
 * @return a value between zero and approximately one
 */
Float LookView::getMaxValue() {
  std::vector<Float>::iterator it;
  Float maxVal(0);
  for(it=integralLookValues.begin(); it!=integralLookValues.end(); it++) 
    if(*it > maxVal)
      maxVal = *it;

  return maxVal;
}

AngleDeg LookView::getAngle(Float& bestValue, AngleDeg min, AngleDeg max) {
  // find best value
  Float minF = getSlice(min);
  Float maxF = getSlice(max);
  if(maxF < minF)
    maxF += slices;

  Float bestSlice = 0;
  bestValue = 0;
  Float val;
  for(int i = (int)floor(minF-0.5); i <= (int)floor(maxF+0.5); i++) {
    val = integralLookValues[i%slices];
    if(val > bestValue) {
      bestValue = val;
      bestSlice = i + 0.5; 
    }
  }
  
  if(bestSlice < minF)
    bestSlice = minF;
  if(bestSlice > maxF)
    bestSlice = maxF;
  
  return AngleDeg(Float(bestSlice) * 360.0 / Float(slices));
}

void LookView::myNewCycle()
{

  // Degrade see-slices
  std::vector<Float>::iterator it;
  for(it = see.begin(); it != see.end(); it++)
    (*it) *= 0.9;

  computeLookValues();
}

void LookView::myUpdateAfterCoachMessage(void)
{
}

void LookView::myUpdateAfterSense(void)
{
}

void LookView::myUpdateAfterSee(void)
{
  const AgentState& agent = worldStateView->getWorldState().getAgent();
  
  Float face = agent.getDirection() + agent.getNeckDirection();

  Float delta(45);
  switch (sensorInterface->getViewWidth())
  {
  case VW_UNKNOWN:
  case VW_NORMAL:
    delta = 45;
    break;
  case VW_WIDE:
    delta = 90;
    break;
  case VW_NARROW:
    delta = 22.5;
    break;
  }

  Float minF = getSlice(face - delta);
  Float maxF = getSlice(face + delta);
  
  Float maxF_index = maxF;    
  if(maxF_index < minF)
    maxF_index += slices;
  
  for(int i = (int)ceil(minF); i <= (int)floor(maxF_index); i++)
      see[i%slices] = 1;

    see[(int)floor(minF)] = ceil(minF) - minF;
    see[(int)ceil(maxF)] = maxF - floor(maxF);

}

void LookView::myUpdateAfterInit(void)
{
}

void LookView::myUpdateAfterHear(void)
{
}

void LookView::myUpdateAfterCommand(void)
{
}

RS_END_NAMESPACE
