/* -*- Mode: C++ -*-
 * $Id: Role.cc,v 1.17 2001/07/31 11:20:35 sigge Exp $
 *
 * IDENTIFICATION
 *
 * $Revision: 1.17 $
 * Created by: Mikael Brnnstrm, Andreas Bckert, Andreas Sigfridsson
 * Last modified by $Author: sigge $ $Date: 2001/07/31 11:20:35 $
 *
 *
 * HISTORY
 *
 * $Log: Role.cc,v $
 * Revision 1.17  2001/07/31 11:20:35  sigge
 * New static passtree.
 *
 * Revision 1.16  2001/07/30 16:00:33  sigge
 * Attract defenders to ball.
 *
 * Revision 1.15  2001/07/29 11:41:52  sigge
 * New team formation.
 *
 * Revision 1.14  2001/07/28 16:22:39  mikbr212
 * Fixed init positions on goal etc
 *
 * Revision 1.13  2001/07/28 14:48:26  sigge
 * Defenders now have floating zones.
 *
 * Revision 1.12  2001/07/28 13:00:19  mikbr212
 * Changed role
 *
 *
 */

/**
 * @file Role.cc
 * Implementation file for the class Role.
 *
 * @par Unit:
 *      Framework
 * $Revision: 1.17 $
 * $Author: sigge $
 * $Date: 2001/07/31 11:20:35 $
 */


#include "Role.h"

RS_BEGIN_NAMESPACE

WorldStateView* Role::worldStateView(NULL);
StatisticsView* Role::statisticsView(NULL);
GameView* Role::gameView(NULL);
ParameterView* Role::parameterView(NULL);

/** Initializes all static stuff (views etc).
 */
void Role::init() {
  static bool hasBeenRun(false);
  if(!hasBeenRun) {
    hasBeenRun = true;
    getView(worldStateView, WORLD_STATE_VIEW_ID);
    getView(statisticsView, STATISTICS_VIEW_ID);
    getView(gameView, GAME_VIEW_ID);
    getView(parameterView, PARAMETER_VIEW_ID);
  } else {
    cout << "Role::init() was called twice" << endl;
  }
}

/** Constructs the agent's role.
 */
Role::Role()
  : uNum(worldStateView->getWorldState().getAgent().getUniformNumber())
{
}

Role::Role(int _uNum) 
  : uNum(_uNum)
{
}

Role::~Role()
{
}

Role::RoleType Role::getRoleType() const {
  return getRoleType(uNum);
}

TVector Role::getHomePosition() const {
  return getHomePosition(uNum);
}

float Role::getZoneRadius() const {
  return getZoneRadius(uNum);
}

TVector Role::getInitPosition() const {
  return getInitPosition(uNum);
}

Role Role::getAgentRole() {
  return Role();
}

Role Role::getPlayerRole(const PlayerState& plr) {
  return Role(plr.getUniformNumber());
}

// static methods
TVector Role::getHomePosition(int uNum) {
  return getHomePosition(uNum, worldStateView->getWorldState().getBall().getPosition());
}

float defY(float ballY, float y)
{
//   float w = 1.0/pow(1.0 + abs(ballY - y), 0.3);
  float w = exp(-pow(abs(ballY - y), 2.0) / 200);
  return w*ballY + (1-w)*y;
}

TVector Role::getHomePosition(int uNum, const TVector& ballPos)
{
  const float defenderBreak = -25;
  float defenderLine;
  if(ballPos.getX() < defenderBreak)
    defenderLine = ballPos.getX();
  else
    defenderLine = (ballPos.getX() - defenderBreak)
      / (52.5 - defenderBreak) * (0 - defenderBreak)
      + defenderBreak;

  switch(uNum)
  {
    // Goal tender
  case 1:
    return TVector(-50,0);

    // Mid defenders
  case 2:
    return TVector(defenderLine+1,defY(ballPos.getY(), -6));
  case 3:
    return TVector(defenderLine+1,defY(ballPos.getY(), 6));
    // Outer defenders
  case 4:
    return TVector(defenderLine-1,defY(ballPos.getY(), -15));
  case 5:
    return TVector(defenderLine-1,defY(ballPos.getY(), 15));
    // Flank defenders
  case 6:
    return TVector(defenderLine+2,defY(ballPos.getY(), -25));
  case 7:
    return TVector(defenderLine+2,defY(ballPos.getY(), 25));
    // Forwards
  case 8:
    return TVector(-1,0);
  case 9:
    return TVector(-5,10);
  case 10:
    return TVector(-1,-20);
  case 11:
    return TVector(-1,25);

  default:
    return TVector(-1,0);
  }
}

TVector Role::getInitPosition(int uNum) {
  return getHomePosition(uNum, TVector(0,0));
}

float Role::getZoneRadius(int uNum) {
  static float radius[11] = {
    10,
    10, 10, 12, 12,
    12, 12, 10,
    10, 10, 10
  };
  if(1 <= uNum && uNum <= 11) 
    return radius[uNum-1];
  return 10;
}

Role::RoleType Role::getRoleType(int uNum) {
  static RoleType roles[11] = {
    rtGoalie,
    rtDefensive,
    rtDefensive,
    rtDefensive,
    rtDefensive,
    rtMid,
    rtMid,
    rtOffensive,
    rtOffensive,
    rtOffensive,
    rtOffensive
  };
  if(1 <= uNum && uNum <= 11) 
    return roles[uNum-1];
  return rtUnknown;
}


RS_END_NAMESPACE
