/* -*- Mode: C++ -*-
 * $Id: StatisticsView.cc,v 1.12 2001/07/31 13:25:03 sigge Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: StatisticsView.cc
 * Unitname: Framework/view
 * $Revision: 1.12 $
 * Last modified by $Author: sigge $ $Date: 2001/07/31 13:25:03 $
 *
 *
 * HISTORY
 *
 * $Log: StatisticsView.cc,v $
 * Revision 1.12  2001/07/31 13:25:03  sigge
 * cyclesSinceKickoff now only resets on real kick offs (not free kicks, kick ins and so on).
 *
 * Revision 1.11  2001/07/29 20:53:57  sigge
 * Added WorldState::getInterceptors and appropriate methods in StatisticsView to accelerate interception calculations.
 *
 * Revision 1.10  2001/07/29 19:17:02  sigge
 * Oops, get the right interceptor in getOurInterceptor.
 *
 * Revision 1.9  2001/07/29 09:43:55  mikbr212
 * Added cyclesSinceOurKickoff
 *
 * Revision 1.8  2001/07/28 14:40:30  sigge
 * Added more intercept cache methods.
 *
 * Revision 1.7  2001/07/26 12:40:58  mikbr212
 * removed PlayerState::getTeam()
 *
 * Revision 1.6  2001/07/19 00:05:47  sigge
 * Fixed step/confidence bug.
 *
 * Revision 1.5  2001/07/12 17:18:02  sigge
 * Added getInterceptionBallState to Statistics view.
 *
 * Revision 1.4  2001/07/06 10:00:38  mikbr212
 * Merged updates and resolved conflicts (a lot)
 *
 */

/**
 * @file StatisticsView.cc
 * Implementation file for the view class StatisticsView.
 *
 * @par Unit:
 *      Framework/view
 * $Revision: 1.12 $
 * $Author: sigge $
 * $Date: 2001/07/31 13:25:03 $
 */

#include "StatisticsView.h"
#include "ParameterView.h"
#include "WorldStateView.h"
#include "GameView.h"

#include "PlayerState.h"
#include <algorithm>

RS_BEGIN_NAMESPACE

/**
 * Constructor. Use getView instead.
 * @see getView
 */
StatisticsView::StatisticsView(const ViewId id)
  : View(id),
    parameterView(NULL),
    worldStateView(NULL),
    gameView(NULL),
    possessionSum(0),
    maxPossessionLength(25),
    tmAboveAgent(0),
    tmBesideAgent(0),
    tmBelowAgent(0),
    opAboveAgent(0),
    opBesideAgent(0),
    opBelowAgent(0),
    tmAboveBall(0),
    tmBesideBall(0),
    tmBelowBall(0),
    opAboveBall(0),
    opBesideBall(0),
    opBelowBall(0),
    interceptionBallStateCacheTime(0),
    cyclesSinceOurKickoff(9999),
    cyclesSinceTheirKickoff(9999)
{
  getView(parameterView, PARAMETER_VIEW_ID);
  getView(worldStateView, WORLD_STATE_VIEW_ID);
  getView(gameView, GAME_VIEW_ID);

  clearLastKicker();
  lastOpGoalPosition = TVector(-40, 0);
}

/** Destructor.
 */
StatisticsView::~StatisticsView()
{
  viewManager->releaseView(PARAMETER_VIEW_ID);
  viewManager->releaseView(WORLD_STATE_VIEW_ID);
  viewManager->releaseView(GAME_VIEW_ID);
}

/** Clears last kicker (team, uniform number and cycle).
 */
void StatisticsView::clearLastKicker() {
  lastKickerTeam = noTeam;
  lastKickerUniformNumber = 0;
  lastKickCycle = 0;
}

/** Updates statistics. Called every cycle.
 */
void StatisticsView::myNewCycle() {
  worldStateView->newCycle(currentUpdateNumber);
  // Check if anyone kicked (could kick) the ball
  checkBallKick();
  // Update the possession value
  updatePossesion();
  // Update ball distance
  updateBallDistance();
  // Update cycle counts
  updateCyclesSinceKickoff();
}

/** Checks if the ball was (could be) kicked by somebody
 * and updates some variables in that case.
 */
void StatisticsView::checkBallKick() {
  unsigned int i;
  const WorldState& ws(worldStateView->getWorldState());
  const BallState& ball(ws.getBall());
  
  // Check opponents
  const std::vector<PlayerState>& opponents(ws.getOpponents());
  for(i = 0; i < opponents.size(); i++) 
    if(opponents[i].isBallKickable(ball)) {
      lastKickerTeam = theirTeam;
      lastKickerUniformNumber = opponents[i].getUniformNumber();
      if(lastKickerUniformNumber < 0 || lastKickerUniformNumber > 11)
	lastKickerUniformNumber = 0;
      lastKickCycle = ws.getGameTime();
      lastOpKickPosition = ball.getPosition();
      return;
    }

  // Check teammates
  const std::vector<PlayerState>& teammates(ws.getTeammates());
  for(i = 0; i < teammates.size(); i++) 
    if(teammates[i].isBallKickable(ball)) {
      lastKickerTeam = ourTeam;
      lastKickerUniformNumber = teammates[i].getUniformNumber();
      if(lastKickerUniformNumber < 0 || lastKickerUniformNumber > 11)
	lastKickerUniformNumber = 0;
      lastKickCycle = ws.getGameTime();
      return;
    }

  // Check agent
  const AgentState& agent(ws.getAgent());
  if(agent.getLastCommandType() == PlayerState::cmdKick) {
    lastKickerTeam = ourTeam;
    lastKickerUniformNumber = agent.getUniformNumber();
    if(lastKickerUniformNumber < 0 || lastKickerUniformNumber > 11)
      lastKickerUniformNumber = 0;
    lastKickCycle = ws.getGameTime() - 1; // The agent kicked the ball last cycle
    return;
  }
}

/** Updates the possession value. Should be called after checkBallKick().
 */
void StatisticsView::updatePossesion() {
  switch(lastKickerTeam) {
  case ourTeam:
    possessionHist.push(1);
    possessionSum++;
    break;
  case theirTeam:
    possessionHist.push(-1);
    possessionSum--;
    break;
  default:
    possessionHist.push(0);
    break;
  }  
  
  if(possessionHist.size() > maxPossessionLength) {
    possessionSum -= possessionHist.front();
    possessionHist.pop();
  }
}

/** Returns a value between -1 and 1. Each cycle that our team owns
 * the ball, the value will will increase, and each cycle their team
 * owns the ball the value will decrease. With "own the ball" it is mean
 * the team that kicked (could kick) it most recently.
 */
float StatisticsView::getPossession() const {
  return ((float)possessionSum) / ((float)possessionHist.size());
}

/** Counts the number of teammates/opponents above, beside and below the
 * agent.
 */
void StatisticsView::updatePlayerCount() {
  const float behindBuffer(1.5);
  const float aboveBuffer(1.5);
  const WorldState& ws(worldStateView->getWorldState());
  const float agentPosX(ws.getAgent().getPosition().getX());
  const float ballPosX(ws.getBall().getPosition().getX());
  std::vector<const PlayerState>::iterator it;
  float curX;

  tmBelowAgent = tmBesideAgent = tmAboveAgent = 0;
  opBelowAgent = opBesideAgent = opAboveAgent = 0;

  tmBelowBall = tmBesideBall = tmAboveBall = 0;
  opBelowBall = opBesideBall = opAboveBall = 0;
  
  const std::vector<PlayerState> tms(ws.getTeammates());
  for(it=tms.begin(); it!=tms.end(); it++) { 
    curX = it->getPosition().getX();
    // count agent relative positions
    if(curX < agentPosX - behindBuffer)
      tmBelowAgent++;
    else if(curX < agentPosX + aboveBuffer)
      tmBesideAgent++;
    else
      tmAboveAgent++;
    // count ball relative positions
    if(curX < ballPosX - behindBuffer)
      tmBelowBall++;
    else if(curX < ballPosX + aboveBuffer)
      tmBesideBall++;
    else
      tmAboveBall++;
  }
    
  const std::vector<PlayerState> ops(ws.getOpponents());
  for(it=ops.begin(); it!=ops.end(); it++) { 
    curX = it->getPosition().getX();
    // count agent relative positions
    if(curX < agentPosX - behindBuffer)
      opBelowAgent++;
    else if(curX < agentPosX + aboveBuffer)
      opBesideAgent++;
    else
      opAboveAgent++;
    // count ball relative positions
    if(curX < ballPosX - behindBuffer)
      opBelowBall++;
    else if(curX < ballPosX + aboveBuffer)
      opBesideBall++;
    else
      opAboveBall++;
  }
}

/** Checks if the player is closest to the ball (in his/her team).
 */
bool StatisticsView::isClosestToBall(const PlayerState& plr) {
  const BallState& ball(worldStateView->getWorldState().getBall());
  float dist((plr.getPosition() - ball.getPosition()).getLength());
  float otherDist = teammateBallDistance; // FIXME: getTeam removed
  return (dist <= otherDist);
}

void StatisticsView::updateBallDistance() {
  const WorldState& ws(worldStateView->getWorldState());
  TVector ballPos(ws.getBall().getPosition());
  float dist;

  // *** teammateBallDistance ***
  const AgentState& agent(ws.getAgent());
  teammateBallDistance = (agent.getPosition() - ballPos).getLength();
 
  std::vector<const PlayerState>::iterator it;
  const std::vector<PlayerState> tms(ws.getTeammates());
  for(it=tms.begin(); it!=tms.end(); it++) { 
    dist = (it->getPosition() - ballPos).getLength();
    if(dist < teammateBallDistance)
      teammateBallDistance = dist;
  }

  // *** opponentBallDistance ***
  opponentBallDistance = 100;

  const std::vector<PlayerState> ops(ws.getOpponents());
  for(it=ops.begin(); it!=ops.end(); it++) { 
    dist = (it->getPosition() - ballPos).getLength();
    if(dist < opponentBallDistance)
      opponentBallDistance = dist;
  }
}

/** Called after the agent hears something (from the referee or another player).
 */
void StatisticsView::myUpdateAfterHear() {
  gameView->updateAfterHear(currentUpdateNumber);
  std::vector<PlayMode> pms(gameView->getPlayModes(NOW));
  std::vector<PlayMode>::iterator pm;
  bool endFor = false;
  for(pm = pms.begin(); pm != pms.end(); pm++) {
    switch(*pm) {
    case PM_THEIR_GOAL:
      lastOpGoalPosition = lastOpKickPosition;
      // Intentionally no break
    case PM_BEFORE_KICK_OFF:
    case PM_HALF_TIME:
    case PM_TIME_UP:
    case PM_TIME_OVER:
    case PM_DROP_BALL:
    case PM_OFFSIDE_KICK:
    case PM_OUR_OFFSIDE_KICK:
    case PM_THEIR_OFFSIDE_KICK:
    case PM_KICK_OFF:
    case PM_OUR_KICK_OFF:
    case PM_THEIR_KICK_OFF:
    case PM_KICK_IN:
    case PM_OUR_KICK_IN:
    case PM_THEIR_KICK_IN:
    case PM_FREE_KICK:
    case PM_OUR_FREE_KICK:
    case PM_THEIR_FREE_KICK:
    case PM_CORNER_KICK:
    case PM_OUR_CORNER_KICK:
    case PM_THEIR_CORNER_KICK:
    case PM_GOAL_KICK:
    case PM_OUR_GOAL_KICK:
    case PM_THEIR_GOAL_KICK:
    case PM_GOAL:
    case PM_OUR_GOAL:
    case PM_GOALIE_GOT_BALL:
    case PM_OUR_GOALIE_GOT_BALL:
    case PM_THEIR_GOALIE_GOT_BALL:
      clearLastKicker();
      endFor = true;
      break;
    default:
      break;
    }
    if(endFor)
      break;
  }
}

/**
 * Gets the first interceptors of each team. Will modify all
 * arguments. If there are no players of one kind, it's interception
 * time will be set to -1.
 * Caches the results
 */
void StatisticsView::getInterceptors(PlayerState& ourInterceptor,
				     int& ourInterceptionTime,
				     PlayerState& theirInterceptor,
				     int& theirInterceptionTime)
{
  updateInterceptorsCache();

  ourInterceptor = ourInterceptorCache;
  ourInterceptionTime = ourInterceptionTimeCache;
  theirInterceptor = theirInterceptorCache;
  theirInterceptionTime = theirInterceptionTimeCache;
}

/**
 * Updates the interceptor cache
 */
void StatisticsView::updateInterceptorsCache()
{
  const WorldState& ws = worldStateView->getWorldState();
  int gameTime = ws.getGameTime();
  if(gameTime != interceptorsCacheTime)
  {
    // Update cache
    interceptorsCacheTime = gameTime;
    ws.getInterceptors(ourInterceptorCache, ourInterceptionTimeCache,
		       theirInterceptorCache, theirInterceptionTimeCache);
  }
  else
  {
    // Already up to date, do nothing
  }
}

// Returns earliest interceptor
// Uses caching methods
const PlayerState& StatisticsView::getInterceptor()
{
  updateInterceptorsCache();
  if(ourInterceptionTimeCache < theirInterceptionTimeCache)
    return ourInterceptorCache;
  else
    return theirInterceptorCache;
}

// Returns their earliest interceptor
// Uses caching methods
const PlayerState& StatisticsView::getTheirInterceptor()
{
  updateInterceptorsCache();
  return theirInterceptorCache;
}

// Returns our earliest interceptor
// Uses caching methods
const PlayerState& StatisticsView::getOurInterceptor()
{
  updateInterceptorsCache();
  return ourInterceptorCache;
}

// Returns time before their first intercept
// Uses caching methods
int StatisticsView::getTheirInterceptionTime()
{
  updateInterceptorsCache();
  return theirInterceptionTimeCache;
}

// Returns time before our first intercept
// Uses caching methods
int StatisticsView::getOurInterceptionTime()
{
  updateInterceptorsCache();
  return ourInterceptionTimeCache;
}

// Returns time before first intercept
// Uses caching methods
int StatisticsView::getInterceptionTime()
{
  return min(getOurInterceptionTime(), getTheirInterceptionTime());
}

// Returns predicted ballstate at first intercept
// Caches the result
const BallState& StatisticsView::getInterceptionBallState()
{
  if(worldStateView->getWorldState().getGameTime()
     == interceptionBallStateCacheTime)
    return interceptionBallState;

  const WorldState& ws = worldStateView->getWorldState();
  const BallState& ball = ws.getBall();
  int interceptionTime = getInterceptionTime();

  interceptionBallState = ball;
  interceptionBallState.step(interceptionTime);

  return interceptionBallState;
}
/** Returns the number of cycles since our kickoff.
 */
int StatisticsView::getCyclesSinceOurKickoff() const {
  return cyclesSinceOurKickoff;
}

/** Returns the number of cycles since their kickoff.
 */
int StatisticsView::getCyclesSinceTheirKickoff() const {
  return cyclesSinceTheirKickoff;
}

void StatisticsView::updateCyclesSinceKickoff() {
  cyclesSinceOurKickoff++;
  cyclesSinceTheirKickoff++;

  vector<PlayMode> playModes = gameView->getPlayModes(NOW);

  if(find(playModes.begin(), playModes.end(), PM_OUR_KICK_OFF)
     != playModes.end())
    cyclesSinceOurKickoff = 0;
  else if(find(playModes.begin(), playModes.end(), PM_THEIR_KICK_OFF)
	  != playModes.end())
    cyclesSinceTheirKickoff = 0;
}


RS_END_NAMESPACE


/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler or debugger does not understand the keyword inline then
 * include the inline definitions here, otherwise include them in the
 * declaration file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifndef RS_USE_INLINE
#  include "StatisticsView.icc"
#endif

