/* -*- Mode: C++ -*-
 * $Id: WorldState.h,v 1.29 2001/07/29 20:53:57 sigge Exp $
 * 
 * HISTORY 
 *
 * $Log: WorldState.h,v $
 * Revision 1.29  2001/07/29 20:53:57  sigge
 * Added WorldState::getInterceptors and appropriate methods in StatisticsView to accelerate interception calculations.
 *
 * Revision 1.28  2001/07/28 13:39:13  andbo928
 * Removed a ton of shadowplayers.
 *
 * Revision 1.27  2001/07/27 13:02:32  mikbr212
 * Added stepR
 *
 * Revision 1.26  2001/07/26 18:59:44  mikbr212
 * Added modelling, but it causes a segmentation fault - therefore disabled
 *
 * Revision 1.25  2001/07/26 13:47:53  sigge
 * Added offside helper functions.
 *
 * Revision 1.24  2001/07/06 20:48:47  mikbr212
 * Added getTeam
 *
 * Revision 1.23  2001/07/06 14:27:46  andbo928
 * Fixed field-mirroring.
 *
 * Revision 1.22  2001/07/06 13:02:32  sigge
 * Removed IntentionDecision from various places.
 *
 * Revision 1.21  2001/07/06 10:00:39  mikbr212
 * Merged updates and resolved conflicts (a lot)
 *
 * Revision 1.20  2001/07/05 17:12:24  andbo928
 * Added transformation code.
 *
 * Revision 1.19  2001/07/02 14:44:27  andbo928
 * Added public keyword to inheritance.
 *
 * Revision 1.18  2001/07/01 12:39:53  andbo928
 * Added new limitations imposed by the new server in simulation. Added doChangeViewFoo methods.
 *
 * Revision 1.17  2001/06/19 14:03:35  mikbr212
 * Added a shortcut: setIntention(Intention*) to agent.setIntention(Intention*)
 *
 * Revision 1.16  2001/06/05 20:33:53  mikbr212
 * Fixed doxygen-stubs for all classes
 *
 */

/**
 * @file WorldState.h
 * Definition file for the class WorldState.
 *
 * @par Unit:
 *      Framework/view
 * $Revision: 1.29 $
 * $Author: sigge $
 * $Date: 2001/07/29 20:53:57 $
 */

#ifndef WORLDSTATE_H
#define WORLDSTATE_H

#include <vector>
#include <list>
#include "library_types.h"
#include "ViewManager.h"
#include "ParameterView.h"
#include "view_constants.h"
#include "PlayerObject.h"
#include "AgentObject.h"
#include "PlayerState.h"
#include "BallState.h"
#include "Step.h"

//  #include "Intention.h"

RS_BEGIN_NAMESPACE

class Intention;

/**
 * @class WorldState WorldState.h
 * Describes a state in the 'world'. 
 * This includes:
 *  The ball, the agent, the teammates and the opponents.
 */

class WorldState : public State
{
public:
  // Nested types.

private:
  // Private methods/attributes.

  // Help method.
  void mergePlayers(vector<PlayerState>& mine,
		    const vector<PlayerState>& other);
  void fixConsitency(list<PlayerState>& players);
/*    void WorldState::insertPlayer(const PlayerState& plr, */
/*  				vector<PlayerState>& matchList, */
/*  				vector<PlayerState>& resultList); */
  vector<PlayerState>::iterator
  findMatch(PlayerState plr, vector<PlayerState>& matchList);

  double matchPlayer(const PlayerState& plr1,
		     const PlayerState& plr2);
  void clearCone(TVector point, float lDir, float rDir);


protected:
  // Protected methods/attributes.
  vector<PlayerState> opponents;
  vector<PlayerState> teammates;
  BallState ball;
  AgentState agent;

  int gameTime;
  Team team;

  void framingCompensate(const WorldState& other);
  void checkCollisions();
  void checkCollisions(PlayerState& plr);
  void checkCollisions(BallState& ball);

  void resolveCollisionHelp(PlayerState& plr1,
			    const PlayerState& plr2);
  void resolveCollision(PlayerState& plr1,
			PlayerState& plr2);
  void resolveCollision(PlayerState& plr,
			BallState& ball);
  
  void changeFrame(PlayerState& plr,
		   const TVector& frameDelta,
		   const float frameDirChange,
		   const TVector& frameOrigin);
  void changeFrame(BallState& ball,
		   const TVector& frameDelta,
		   const float frameDirChange,
		   const TVector& frameOrigin);

  static PlayerState modelPlayer(WorldState& world);
  void model();

  void mirror(PlayerState& plr);
  void swapSides();

public:
  // Public methods/attributes.
  const vector<PlayerState>& getOpponents() const;
  void setOpponents(const vector<PlayerState>& vec);

  const vector<PlayerState>& getTeammates() const;
  void setTeammates(const vector<PlayerState>& vec);

  const BallState& getBall() const;
  void setBall(const BallState& other);

  const AgentState& getAgent() const;
  void setAgent(const AgentState& other);

  Team getTeam() const;

  int getGameTime() const;
  void setGameTime(int gt);

  const PlayerState& getTheirInterceptor() const;
  const PlayerState& getOurInterceptor() const;
  void getInterceptors(PlayerState& ourInterceptor,
		       int& ourInterceptionTime,
		       PlayerState& theirInterceptor,
		       int& theirInterceptionTime) const;

  void merge(const WorldState& other);
  void merge(TVector obsPoint, float lDir, float rDir, 
	     const WorldState& other);

  WorldState();
  virtual ~WorldState();

  void reduceC(float conf);
  void reduceR(float radius);
  void reduceCR(float conf, float radius);

  // Convinient defines.
  virtual bool doKick(const TVector& v);
  virtual bool doKickToSpeed(const TVector& speed);
  virtual bool doKickToPosition(const TVector& pos);

  virtual bool doDash(float power);
  virtual bool doDashDistance(float dist);

  virtual bool doTurn(float angle);
  virtual bool doActualTurn(float angle);
  virtual bool doTurnTo(float angle);
  virtual bool doTurnTowards(const TVector& position);
  virtual bool doTurnTowardsBall();

  virtual bool doTurnNeck(float angle);
  virtual bool doTurnNeckTo(float dir);
  virtual bool doTurnNeckTowards(const TVector& position);
  virtual bool doTurnNeckTowardsBall();

  virtual bool doNull();

  virtual bool doChangeViewWide();
  virtual bool doChangeViewNormal();
  virtual bool doChangeViewNarrow();
  virtual bool doChangeViewHigh();
  virtual bool doChangeViewLow();

  virtual void setIntention(Intention* intention);  

  void changeAgent(const PlayerState& newAgent);

  void print();

  Step getStep();

  // Public methods/attributes.
  virtual void step();
  virtual void stepR(float stepDistance); 
  virtual void stepR(float stepDistance, const TVector& center);
  // Returns number of chars needed.
  virtual string serialize() const;
  // Returns number of chars consumed.
  virtual string deserialize(const string& buf);

  float getOffsideLine() const;
  bool isPlayerOffside(const PlayerState& player) const;
};

RS_END_NAMESPACE
#endif /* WORLDSTATE_H */
